"""
These functions help you answer one of the most common questions posed in spatial analysis: "What is near what?"

connect_origins_to_destinations measures the travel time or distance between pairs of points.
create_buffers create areas of equal distance from features.
create_drive_time_areas finds areas around locations that can be reached within a time period.
find_nearest identifies those places that are the closest to known locations.
plan_routes determines the best way to route a fleet of vehicles to visit many stops.
"""
import arcgis as _arcgis
from arcgis._impl.common._utils import _date_handler
import arcgis.network as network


def connect_origins_to_destinations(origins_layer,
                                    destinations_layer,
                                    measurement_type="DrivingTime",
                                    origins_layer_route_id_field=None,
                                    destinations_layer_route_id_field=None,
                                    time_of_day=None,
                                    time_zone_for_time_of_day="GeoLocal",
                                    output_name=None,
                                    context=None,
                                    gis=None,
                                    estimate=False,
                                    point_barrier_layer=None,
                                    line_barrier_layer=None,
                                    polygon_barrier_layer=None,
                                    future=False,
                                    route_shape='FollowStreets'):
    """
    .. image:: _static/images/connect_origins_to_destinations/connect_origins_to_destinations.png
    
    The Connect Origins to Destinations task measures the travel time or distance between pairs of points. Using this tool, you can

    * Calculate the total distance or time commuters travel on their home-to-work trips.
    * Measure how far customers are traveling to shop at your stores. Use this information to define your market reach, especially when targeting advertising campaigns or choosing new store locations.
    * Calculate the expected trip mileage for your fleet of vehicles. Afterward, run the Summarize Within tool to report mileage by state or other region.

    You provide starting and ending points, and the tool returns a layer containing route lines, including measurements, between the
    paired origins and destinations.

    ===================================    =========================================================
    **Argument**                           **Description**
    -----------------------------------    ---------------------------------------------------------
    origins_layer                          Required layer. The starting point or points of the routes to be generated. See :ref:`Feature Input<FeatureInput>`.
    -----------------------------------    ---------------------------------------------------------
    destinations_layer                     Required layer. The routes end at points in the destinations layer. See :ref:`Feature Input<FeatureInput>`.
    -----------------------------------    ---------------------------------------------------------
    measurement_type                       Required string. The origins and destinations can be connected by measuring straight-line distance,
                                           or by measuring travel time or travel distance along a street network using various modes of transportation
                                           known as travel modes.

                                           Valid values are a string, StraightLine, which indicates Euclidean distance to be used as distance measure or
                                           a Python dictionary representing settings for a travel mode.

                                           When using a travel mode for the measurement_type, you need to specify a dictionary
                                           containing the settings for a travel mode supported by your organization. The code in the example section below generates
                                           a valid Python dictionary and then passes it as the value for the measurement_type parameter.

                                           Supported travel modes: ['Driving Distance', 'Driving Time', 'Rural Driving Distance', 'Rural Driving Time',
                                           'Trucking Distance', 'Trucking Time', 'Walking Distance', 'Walking Time']

    -----------------------------------    ---------------------------------------------------------
    origins_layer_route_id_field           Optional string. Specify the field in the origins layer containing the IDs that pair origins with destinations.

                                           * The ID values must uniquely identify points in the origins layer.

                                           * Each ID value must also correspond with exactly one route ID value in the destinations layer. Route IDs that match
                                             across the layers create origin-destination pairs, which the tool connects together.

                                           *  Specifying origins_layer_route_id_field is optional when there is exactly one point feature in the origins or
                                              destinations layer. The tool will connect all origins to the one destination or the one origin to all destinations,
                                              depending on which layer contains one point.
    -----------------------------------    ---------------------------------------------------------
    destinations_layer_route_id_field      Optional string. Specify the field in the destinations layer containing the IDs that pair origins with destinations.

                                           * The ID values must uniquely identify points in the destinations layer.

                                           * Each ID value must also correspond with exactly one route ID value in the origins layer. Route IDs that match across the
                                             layers create origin-destination pairs, which the tool connects together.

                                           * Specifying destinations_layer_route_id_field is optional when there is exactly one point feature in the origins or
                                             destinations layer. The tool will connect all origins to the one destination or the one origin to all destinations,
                                             depending on which layer contains one point.
    -----------------------------------    ---------------------------------------------------------
    time_of_day                            Optional datetime.datetime. Specify whether travel times should consider traffic conditions. To use traffic in the analysis,
                                           set measurement_type to a travel mode object whose impedance_attribute_name property is set to travel_time and assign a value
                                           to time_of_day. (A travel mode with other impedance_attribute_name values don't support traffic.) The time_of_day value represents
                                           the time at which travel begins, or departs, from the origin points. The time is specified as datetime.datetime.

                                           The service supports two kinds of traffic: typical and live. Typical traffic references travel speeds that are made up of historical
                                           averages for each five-minute interval spanning a week. Live traffic retrieves speeds from a traffic feed that processes phone probe
                                           records, sensors, and other data sources to record actual travel speeds and predict speeds for the near future.

                                           The `data coverage <http://www.arcgis.com/home/webmap/viewer.html?webmap=b7a893e8e1e04311bd925ea25cb8d7c7>`_ page shows the countries
                                           Esri currently provides traffic data for.

                                           Typical Traffic:

                                           To ensure the task uses typical traffic in locations where it is available, choose a time and day of the week, and then convert the day
                                           of the week to one of the following dates from 1990:

                                           * Monday - 1/1/1990
                                           * Tuesday - 1/2/1990
                                           * Wednesday - 1/3/1990
                                           * Thursday - 1/4/1990
                                           * Friday - 1/5/1990
                                           * Saturday - 1/6/1990
                                           * Sunday - 1/7/1990
                                           Set the time and date as datetime.datetime.

                                           For example, to solve for 1:03 p.m. on Thursdays, set the time and date to 1:03 p.m., 4 January 1990; and convert to
                                           datetime eg. datetime.datetime(1990, 1, 4, 1, 3).

                                           Live Traffic:

                                           To use live traffic when and where it is available, choose a time and date and convert to datetime.

                                           Esri saves live traffic data for 12 hours and references predictive data extending 12 hours into the future. If the time and date you
                                           specify for this parameter is outside the 24-hour time window, or the travel time in the analysis continues past the predictive data window, the task falls back to typical traffic speeds.

                                           Examples:
                                           from datetime import datetime

                                           * "time_of_day": datetime(1990, 1, 4, 1, 3) # 13:03, 4 January 1990. Typical traffic on Thursdays at 1:03 p.m.
                                           * "time_of_day": datetime(1990, 1, 7, 17, 0) # 17:00, 7 January 1990. Typical traffic on Sundays at 5:00 p.m.
                                           * "time_of_day": datetime(2014, 10, 22, 8, 0) # 8:00, 22 October 2014. If the current time is between 8:00 p.m., 21 Oct. 2014 and 8:00 p.m., 22 Oct. 2014,
                                             live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.
                                           * "time_of_day": datetime(2015, 3, 18, 10, 20) # 10:20, 18 March 2015. If the current time is between 10:20 p.m., 17 Mar. 2015 and 10:20 p.m., 18 Mar. 2015,
                                             live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.

    -----------------------------------    ---------------------------------------------------------
    time_zone_for_time_of_day              Optional string. Specify the time zone or zones of the timeOfDay parameter.
                                           Choice list: ['GeoLocal', 'UTC']

                                           GeoLocal-refers to the time zone in which the originsLayer points are located.

                                           UTC-refers to Coordinated Universal Time.
    -----------------------------------    ---------------------------------------------------------
    include_route_layers                   Optional Boolean. When include_route_layers is set to True, each route from the result is also saved as a route layer item. A route layer includes all the
                                           information for a particular route such as the stops assigned to the route as well as the travel directions. Creating route layers is useful if you want
                                           to share individual routes with other members in your organization. The route layers use the output feature service name provided in the outputName
                                           parameter as a prefix and the route name generated as part of the analysis is added to create a unique name for each route layer.

                                           Caution: Route layers cannot be created when the output is a feature collection. The task will raise an error if output_name is not specified
                                           (which indicates feature collection output) and include_route_layers is True.

                                           The maximum number of route layers that can be created is 1,000. If the result contains more than 1,000 routes and include_route_layers is True,
                                           the task will only create the output feature service.
    -----------------------------------    ---------------------------------------------------------
    output_name                            Optional string. If provided, the task will create a feature layer of the results. You define the name of the layer. If output_name is not supplied,
                                           the task will return a feature collection.
    -----------------------------------    ---------------------------------------------------------
    context                                Optional string. Additional settings such as processing extent and output spatial reference. For calculate_density, there are two settings.

                                           #. Extent (extent)-a bounding box that defines the analysis area. Only those points in the origins_layer and destinations_layer that intersect the
                                              bounding box will be analyzed.
                                           #. Output Spatial Reference (outSR)-If the output is a feature service, the spatial reference will be the same as originsLayer. Setting outSR for feature services has no effect.
                                              If the output is a feature collection, the features will be in the spatial reference of the outSRvalue or the spatial reference of originsLayer when outSR is not specified.
    -----------------------------------    ---------------------------------------------------------
    gis                                    Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -----------------------------------    ---------------------------------------------------------
    estimate                               Optional Boolean. Is True, the number of credits needed to run the operation will be returned as a float.
    -----------------------------------    ---------------------------------------------------------
    point_barrier_layer                    Optional layer. Specify one or more point features that act as temporary restrictions (in other words, barriers) when traveling on the underlying streets.

                                           A point barrier can model a fallen tree, an accident, a downed electrical line, or anything that completely blocks traffic at a specific position along
                                           the street. Travel is permitted on the street but not through the barrier. See :ref:`Feature Input<FeatureInput>`.
    -----------------------------------    ---------------------------------------------------------
    line_barrier_layer                     Optional layer. Specify one or more line features that prohibit travel anywhere the lines intersect the streets.

                                           A line barrier prohibits travel anywhere the barrier intersects the streets. For example, a parade or protest that blocks traffic across several street
                                           segments can be modeled with a line barrier. See :ref:`Feature Input<FeatureInput>`.
    -----------------------------------    ---------------------------------------------------------
    polygon_barrier_layer                  Optional string. Specify one or more polygon features that completely restrict travel on the streets intersected by the polygons.

                                           One use of this type of barrier is to model floods covering areas of the street network and making road travel there impossible. See :ref:`Feature Input<FeatureInput>`.
    -----------------------------------    ---------------------------------------------------------
    future                                 Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    -----------------------------------    ---------------------------------------------------------
    route_shape                            Optional String. Specify the shape of the route that connects each origin to it's destination when using a travel mode.

                                           Values: FollowStreets or StraightLine

                                           Default: FollowStreets

                                              + FollowStreets - The shape is based on the underlying street network. This option is best when you want to generate the routes between origins and destinations. This is the default value when using a travel mode.
                                              + StraightLine - The shape is a straight line connecting the origin-destination pair. This option is best when you want to generate spider diagrams or desire lines (for example, to show which stores customers are visiting). This is the default value when not using a travel mode.
                                           
                                           The best route between an origin and it's matched destination is always calculated based on the travel mode, regardless of which route shape is chosen.
    ===================================    =========================================================


    :returns: dict with the following keys:

        "routes_layer" : layer (FeatureCollection)

        "unassigned_origins_layer" : layer (FeatureCollection)

        "unassigned_destinations_layer" : layer (FeatureCollection)

    .. code-block:: python

        USAGE EXAMPLE: To retrieve trvel modes and run connect_origins_to_destinations tool.

        This example creates route between esri regional offices to esri headquarter.

        routes =  connect_origins_to_destinations(origins_layer=esri_regional,
                                         destinations_layer=dest_layer,
                                         measurement_type='Rural Driving Distance',
                                         time_of_day=datetime(1990, 1, 4, 1, 3),
                                         output_name="routes_from_offices_to_hq")
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    
    if isinstance(measurement_type, str):
        route_service = network.RouteLayer(gis.properties.helperServices.route.url, gis=gis)
        travelmodes = route_service.retrieve_travel_modes()

        for tm in travelmodes['supportedTravelModes']:
            if tm['name'] == measurement_type:
                tm = [i for i in route_service.retrieve_travel_modes()['supportedTravelModes'] if i['name'] == measurement_type][0]
                measurement_type = tm

    return gis._tools.featureanalysis.connect_origins_to_destinations(
        origins_layer,
        destinations_layer,
        measurement_type,
        origins_layer_route_id_field,
        destinations_layer_route_id_field,
        _date_handler(time_of_day),
        time_zone_for_time_of_day,
        output_name,
        context,
        estimate=estimate, 
        route_shape=route_shape,
        future=future)


def create_buffers(
        input_layer,
        distances=[],
        field=None,
        units="Meters",
        dissolve_type="None",
        ring_type="Disks",
        side_type="Full",
        end_type="Round",
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        future=False):
    """
    .. image:: _static/images/create_buffers/create_buffers.png

    .. |Disks| image:: _static/images/create_buffers/Disks.png
    .. |Dissolve| image:: _static/images/create_buffers/Dissolve.png
    .. |Flat| image:: _static/images/create_buffers/Flat.png
    .. |Full| image:: _static/images/create_buffers/Full.png
    .. |Left| image:: _static/images/create_buffers/Left.png
    .. |None| image:: _static/images/create_buffers/None.png
    .. |Outside| image:: _static/images/create_buffers/Outside.png
    .. |Right| image:: _static/images/create_buffers/Right.png
    .. |Rings| image:: _static/images/create_buffers/Rings.png
    .. |Round| image:: _static/images/create_buffers/Round.png
    .. |Unspecified| image:: _static/images/create_buffers/Unspecified.png

    The ``create_buffers`` task creates polygons that cover a given distance from a point,
    line, or polygon feature. Buffers are typically used to create areas that can be
    further analyzed using a tool such as ``overlay_layers``. For example, if the question
    is "What buildings are within one mile of the school?", the answer can be found by
    creating a one-mile buffer around the school and overlaying the buffer with the layer
    containing building footprints. The end result is a layer of those buildings within
    one mile of the school.

    =========================    =========================================================
    **Parameter**                **Description**
    -------------------------    ---------------------------------------------------------
    input_layer                  Required point, line or polygon feature layer. The input features to be buffered. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    distances                    Optional list of floats to buffer the input features. The distance(s) that will be buffered. You must supply values
                                 for either the ``distances`` or ``field`` parameter. You can enter a single distance value or multiple values.
                                 The units of the distance values is suppied by the units parameter.
    -------------------------    ---------------------------------------------------------
    field                        Optional string. A field on the ``input_layer`` containing a buffer distance. Buffers will be created using field values.
                                 Unlike the ``distances`` parameter, multiple distances are not supported on field input.
    -------------------------    ---------------------------------------------------------
    units                        Optional string. The linear unit to be used with the distance value(s) specified in distances or contained in the field value.

                                 Choice list: ['Meters', 'Kilometers', 'Feet', 'Miles', 'NauticalMiles', 'Yards']

                                 The default is 'Meters'.
    -------------------------    ---------------------------------------------------------
    dissolve_type                Optional string. Determines how overlapping buffers are processed.

                                 Choice list: ['None', 'Dissolve']

                                 +------------+---------------------------------------------------------------------------------+
                                 | |None|     | ``None``-Overlapping areas are kept. This is the default.                       |
                                 +------------+---------------------------------------------------------------------------------+
                                 | |Dissolve| | ``Dissolve``-Overlapping areas are combined.                                    |
                                 +------------+---------------------------------------------------------------------------------+

    -------------------------    ---------------------------------------------------------
    ring_type                    Optional string. Determines how multiple-distance buffers are processed.

                                 Choice list: ['Disks', 'Rings']

                                 +-----------+--------------------------------------------------------------------------------------------------+
                                 | |Disks|   | ``Disks``-buffers are concentric and will overlap. For example, if your distances are 10 and 14, |
                                 |           | the result will be two buffers, one from 0 to 10 and one from 0 to 14. This is the default.      |
                                 +-----------+--------------------------------------------------------------------------------------------------+
                                 | |Rings|   | ``Rings`` buffers will not overlap. For example, if your distances are 10 and 14, the result will|
                                 |           | be two buffers, one from 0 to 10 and one from 10 to 14.                                          |
                                 +-----------+--------------------------------------------------------------------------------------------------+

    -------------------------    ---------------------------------------------------------
    side_type                    Optional string. When buffering line features, you can choose which side of the line to buffer.

                                 Typically, you choose both sides (Full, which is the default). Left and right are determined as
                                 if you were walking from the first x,y coordinate of the line (the start coordinate) to the last
                                 x,y coordinate of the line (the end coordinate). Choosing left or right usually means you know
                                 that your line features were created and stored in a particular direction (for example, upstream
                                 or downstream in a river network).

                                 When buffering polygon features, you can choose whether the buffer includes or excludes the polygon
                                 being buffered.

                                 Choice list: ['Full', 'Left', 'Right', 'Outside']

                                 +---------------+----------------------------------------------------------------------------------------------------+
                                 | |Full|        | ``Full``-both sides of the line will be buffered. This is the default for line featuress.          |
                                 |               |                                                                                                    |
                                 +---------------+----------------------------------------------------------------------------------------------------+
                                 | |Left|        | ``Left``-only the right side of the line will be buffered.                                         |
                                 +---------------+----------------------------------------------------------------------------------------------------+
                                 | |Right|       | ``Right``-only the right side of the line will be buffered.                                        |
                                 +---------------+----------------------------------------------------------------------------------------------------+
                                 | |Outside|     | ``Outside`` when buffering a polygon, the polygon being buffered is excluded in the result buffer. |
                                 +---------------+----------------------------------------------------------------------------------------------------+
                                 | |Unspecified| | If ``side_type`` not supplied, the polygon being buffered is included in the result buffer.        |
                                 |               | This is the  default for polygon features.                                                         |
                                 +---------------+----------------------------------------------------------------------------------------------------+

    -------------------------    ---------------------------------------------------------
    end_type                     Optional string. The shape of the buffer at the end of line input features. This parameter is not
                                 valid for polygon input features. At the ends of lines the buffer can be rounded (Round) or be
                                 straight across (Flat).

                                 Choice list: ['Round', 'Flat']

                                 +---------+-------------------------------------------------------------------------------+
                                 | |Round| | ``Round``-buffers will be rounded at the ends of lines. This is the default.  |
                                 +---------+-------------------------------------------------------------------------------+
                                 | |Flat|  | ``Flat``-buffers will be flat at the ends of lines.                           |
                                 +---------+-------------------------------------------------------------------------------+

    -------------------------    ---------------------------------------------------------
    output_name                  Optional string. Output feature service name. If not provided, a feature collection is returned.
    -------------------------    ---------------------------------------------------------
    context                      Optional dict. Context contains additional settings that affect task execution. For ``create_buffers``, there are two settings.

                                 #. Extent (``extent``)-a bounding box that defines the analysis area. Only those points in the ``input_layer``
                                    that intersect the bounding box will be analyzed.

                                 #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    -------------------------    ---------------------------------------------------------
    gis                          Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------    ---------------------------------------------------------
    estimate                     Optional boolean. If True, the estimated number of credits required to run the operation will be returned.
    -------------------------    ---------------------------------------------------------
    future                       Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =========================    =========================================================

    :returns: result_layer : feature layer Item if output_name is specified, else Feature Collection.


    .. code-block:: python

        USAGE EXAMPLE: To create 5 mile buffer around US parks, within the specified extent.

        polygon_lyr_buffer = create_buffers(input_layer=parks_lyr,
                                 distances=[5],
                                 units='Miles',
                                 ring_type='Rings',
                                 end_type='Flat',
                                 output_name='create_buffers',
                                 context={"extent":{"xmin":-12555831.656684224,"ymin":5698027.566358956,"xmax":-11835489.102124758,"ymax":6104672.556836072,"spatialReference":{"wkid":102100,"latestWkid":3857}}})
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.create_buffers(input_layer,
                                                     distances,
                                                     field,
                                                     units,
                                                     dissolve_type,
                                                     ring_type,
                                                     side_type,
                                                     end_type,
                                                     output_name,
                                                     context,
                                                     estimate=estimate,
                                                     future=future)


def create_drive_time_areas(input_layer,
                            break_values=[5, 10, 15],
                            break_units="Minutes",
                            travel_mode="Driving",
                            overlap_policy="Overlap",
                            time_of_day=None,
                            time_zone_for_time_of_day="GeoLocal",
                            output_name=None,
                            context=None,
                            gis=None,
                            estimate=False,
                            point_barrier_layer=None,
                            line_barrier_layer=None,
                            polygon_barrier_layer=None,
                            future=False,
                            travel_direction="AwayFromFacility",
                            show_holes=False):
    """
    .. image:: _static/images/create_drive_time_areas/create_drive_time_areas.png

    .. |Overlap| image:: _static/images/create_drive_time_areas/Overlap.png
    .. |Dissolve| image:: _static/images/create_drive_time_areas/Dissolve.png
    .. |Split| image:: _static/images/create_drive_time_areas/Split.png

    The ``create_drive_time_areas`` method creates areas that can be reached within a
    given drive time or drive distance. It can help you answer questions such as:

    * How far can I drive from here in five minutes?
    * What areas are covered within a three-mile drive distance of my stores?
    * What areas are within four minutes of our fire stations?

    =========================    =========================================================
    **Parameter**                **Description**
    -------------------------    ---------------------------------------------------------
    input_layer                  Required point feature layer. The points around which travel areas
                                 based on a mode of transportation will be drawn.
                                 See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    travel_mode                  Optional string or dict. Specify the mode of transportation for the analysis.

                                 Choice list: ['Driving Distance', 'Driving Time', 'Rural Driving Distance', 'Rural Driving Time', 'Trucking Distance', 'Trucking Time', 'Walking Distance', 'Walking Time']

                                 The default is 'Driving Time'.
    -------------------------    ---------------------------------------------------------
    break_values                 Optional list of floats. The size of the polygons to create.
                                 The units for break_values is specified with the break_units parameter.

                                 By setting many unique values in the list, polygons of different sizes are generated around each input location.

                                 The default is [5, 10, 15].
    -------------------------    ---------------------------------------------------------
    break_units                  Optional string. The units of the break_values parameter.

                                 To create areas showing how far you can go along roads or walkways within a given time, specify a time unit.
                                 Alternatively, specify a distance unit to generate areas bounded by a maximum travel distance.

                                 When the travel_mode is time based, a time unit should be specified for the break_units. When the
                                 travel_mode is distance based, a distance unit should be specified for the break_units.

                                 Choice list: ['Seconds', 'Minutes', Hours', 'Feet', 'Meters', 'Kilometers', 'Feet', 'Miles', 'Yards']

                                 The default is 'Minutes'.
    -------------------------    ---------------------------------------------------------
    overlap_policy               Optional string. Determines how overlapping areas are processed.

                                 Choice list: ['Overlap', 'Dissolve', 'Split']

                                 +---------------+-----------------------------------------------------------------------------------------------------+
                                 | |Overlap|     | ``Overlap``-Overlapping areas are kept. This is the default.                                        |
                                 +---------------+-----------------------------------------------------------------------------------------------------+
                                 | |Dissolve|    | ``Dissolve``-Overlapping areas are combined by break value. Because the areas are dissolved,        |
                                 |               | use this option when you need to know the areas that can be reached within a                        |
                                 |               | given time or distance, but you don't need to know which input points are nearest.                  |
                                 +---------------+-----------------------------------------------------------------------------------------------------+
                                 | |Split|       | ``Split``-Overlapping areas are split in the middle. Use this option when you need to know          |
                                 |               | the one nearest input location to the covered area.                                                 |
                                 +---------------+-----------------------------------------------------------------------------------------------------+

                                 The default is 'Overlap'

    -------------------------    ---------------------------------------------------------
    time_of_day                  Optional datetime.datetime. Specify whether travel times should consider traffic conditions. To use traffic in the analysis,
                                 set measurement_type to a travel mode object whose impedance_attribute_name property is set to travel_time and assign a value
                                 to time_of_day. (A travel mode with other impedance_attribute_name values don't support traffic.) The time_of_day value represents
                                 the time at which travel begins, or departs, from the origin points. The time is specified as datetime.datetime.

                                 The service supports two kinds of traffic: typical and live. Typical traffic references travel speeds that are made up of historical
                                 averages for each five-minute interval spanning a week. Live traffic retrieves speeds from a traffic feed that processes phone probe
                                 records, sensors, and other data sources to record actual travel speeds and predict speeds for the near future.

                                 The `data coverage <http://www.arcgis.com/home/webmap/viewer.html?webmap=b7a893e8e1e04311bd925ea25cb8d7c7>`_ page shows the countries
                                 Esri currently provides traffic data for.

                                 Typical Traffic:

                                 To ensure the task uses typical traffic in locations where it is available, choose a time and day of the week, and then convert the day
                                 of the week to one of the following dates from 1990:

                                 * Monday - 1/1/1990
                                 * Tuesday - 1/2/1990
                                 * Wednesday - 1/3/1990
                                 * Thursday - 1/4/1990
                                 * Friday - 1/5/1990
                                 * Saturday - 1/6/1990
                                 * Sunday - 1/7/1990
                                 Set the time and date as datetime.datetime.

                                 For example, to solve for 1:03 p.m. on Thursdays, set the time and date to 1:03 p.m., 4 January 1990; and convert to
                                 datetime eg. datetime.datetime(1990, 1, 4, 1, 3).

                                 Live Traffic:

                                 To use live traffic when and where it is available, choose a time and date and convert to datetime.

                                 Esri saves live traffic data for 12 hours and references predictive data extending 12 hours into the future. If the time and date you
                                 specify for this parameter is outside the 24-hour time window, or the travel time in the analysis continues past the predictive data window, the task falls back to typical traffic speeds.

                                 Examples:
                                 from datetime import datetime

                                 * "time_of_day": datetime(1990, 1, 4, 1, 3) # 13:03, 4 January 1990. Typical traffic on Thursdays at 1:03 p.m.
                                 * "time_of_day": datetime(1990, 1, 7, 17, 0) # 17:00, 7 January 1990. Typical traffic on Sundays at 5:00 p.m.
                                 * "time_of_day": datetime(2014, 10, 22, 8, 0) # 8:00, 22 October 2014. If the current time is between 8:00 p.m., 21 Oct. 2014 and 8:00 p.m., 22 Oct. 2014,
                                   live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.
                                 * "time_of_day": datetime(2015, 3, 18, 10, 20) # 10:20, 18 March 2015. If the current time is between 10:20 p.m., 17 Mar. 2015 and 10:20 p.m., 18 Mar. 2015,
                                   live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.
    -------------------------    ---------------------------------------------------------
    time_zone_for_time_of_day    Optional string. Specify the time zone or zones of the time_of_day parameter.

                                 Choice list: ['GeoLocal', 'UTC']

                                 GeoLocal-refers to the time zone in which the originsLayer points are located.

                                 UTC-refers to Coordinated Universal Time.

                                 The default is 'GeoLocal'.
    -------------------------    ---------------------------------------------------------
    output_name                  Optional string. Output feature service name. If not provided, a feature collection is returned.
    -------------------------    ---------------------------------------------------------
    context                      Optional dict. Context contains additional settings that affect task execution. For ``create_drive_time_areas``, there are two settings.

                                 #. Extent (``extent``)-a bounding box that defines the analysis area. Only those points in the ``input_layer``
                                    that intersect the bounding box will be analyzed.

                                 #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    -------------------------    ---------------------------------------------------------
    gis                          Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------    ---------------------------------------------------------
    estimate                     Optional boolean. If True, the estimated number of credits required to run the operation will be returned.
    -------------------------    ---------------------------------------------------------
    point_barrier_layer          Optional layer. Specify one or more point features that act as temporary restrictions (in other words, barriers)
                                 when traveling on the underlying streets.

                                 A point barrier can model a fallen tree, an accident, a downed electrical line, or anything that completely blocks
                                 traffic at a specific position along the street. Travel is permitted on the street but not through the barrier. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    line_barrier_layer           Optional layer. Specify one or more line features that prohibit travel anywhere the lines intersect the streets.

                                 A line barrier prohibits travel anywhere the barrier intersects the streets. For example, a parade or protest that blocks traffic across several street
                                 segments can be modeled with a line barrier. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    polygon_barrier_layer        Optional string. Specify one or more polygon features that completely restrict travel on the streets intersected by the polygons.

                                 One use of this type of barrier is to model floods covering areas of the street network and making road travel there impossible. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    future                       Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    -------------------------    ---------------------------------------------------------
    travel_direction             Optiona String. Specify whether the direction of travel used to generate the travel areas is toward or away from the input locations.

                                 Values: AwayFromFacility or TowardsFacility

                                 The travel direction can influence how the areas are generated. CreateDriveTimeAreas will obey one-way streets, avoid illegal turns, and follow other rules based on the direction of travel. You should select the direction of travel based on the type of input locations and the context of your analysis. For example, the drive-time area for a pizza delivery store should be created away from the facility, whereas the drive-time area for a hospital should be created toward the facility.
    -------------------------    ---------------------------------------------------------
    show_holes                   Optional boolean. When set to true, the output areas will include holes if some streets couldn't be reached without exceeding the cutoff or due to travel restrictions imposed by the travel mode.
    =========================    =========================================================

    :returns: result_layer : feature layer Item if output_name is specified, else Feature Collection.


    .. code-block:: python

        USAGE EXAMPLE: To create drive time areas around USA airports, within the specified extent.

        target_area4 = create_drive_time_areas(airport_lyr,
                                       break_values=[2, 4],
                                       break_units='Hours',
                                       travel_mode='Trucking Time',
                                       overlap_policy='Split',
                                       time_of_day=datetime(2019, 5, 13, 7, 52),
                                       output_name='create_drive_time_areas',
                                       context={"extent":{"xmin":-11134400.655784884,"ymin":3368261.7800108367,"xmax":-10682810.692676282,"ymax":3630899.409198575,"spatialReference":{"wkid":102100,"latestWkid":3857}}}) """

    gis = _arcgis.env.active_gis if gis is None else gis
    if isinstance(travel_mode, str):
        route_service = network.RouteLayer(gis.properties.helperServices.route.url, gis=gis)
        travelmodes = route_service.retrieve_travel_modes()
        for tm in travelmodes['supportedTravelModes']:
            if tm['name'] == travel_mode:
                tm = [i for i in route_service.retrieve_travel_modes()['supportedTravelModes'] if i['name'] == travel_mode][0]
                travel_mode = tm

    return gis._tools.featureanalysis.create_drive_time_areas(
        input_layer,
        break_values,
        break_units,
        travel_mode,
        overlap_policy,
        _date_handler(time_of_day),
        time_zone_for_time_of_day,
        output_name,
        context,
        estimate=estimate,
        point_barrier_layer=point_barrier_layer,
        line_barrier_layer=line_barrier_layer,
        polygon_barrier_layer=polygon_barrier_layer, future=future,
        travel_direction=travel_direction, 
        show_holes=show_holes)
        


def find_nearest(
        analysis_layer,
        near_layer,
        measurement_type="StraightLine",
        max_count=100,
        search_cutoff=2147483647,
        search_cutoff_units=None,
        time_of_day=None,
        time_zone_for_time_of_day="GeoLocal",
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        include_route_layers=None,
        point_barrier_layer=None,
        line_barrier_layer=None,
        polygon_barrier_layer=None,
        future=False):
    """
    .. image:: _static/images/find_nearest/find_nearest.png

    The ``find_nearest`` method measures the straight-line distance, driving distance, or driving time from
    features in the analysis layer to features in the near layer, and copies the nearest features in the near
    layer to a new layer. Connecting lines showing the measured path are returned as well. ``find_nearest`` also
    reports the measurement and relative rank of each nearest feature. There are options to limit the number
    of nearest features to find or the search range in which to find them. The results from this method can help
    you answer the following kinds of questions:

    * What is the nearest park from here?
    * Which hospital can I reach in the shortest drive time? And how long would the trip take on a Tuesday at 5:30 p.m. during rush hour?
    * What are the road distances between major European cities?
    * Which of these patients reside within two miles of these chemical plants?

    Find Nearest returns a layer containing the nearest features and a line layer that links the start locations to their nearest locations.
    The connecting line layer contains information about the start and nearest locations and the distances between.

    =========================    =========================================================
    **Parameter**                **Description**
    -------------------------    ---------------------------------------------------------
    analysis_layer               Required layer. The features from which the nearest locations are found. This layer can have point, line, or polygon features. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    near_layer                   Required layer. The nearest features are chosen from this layer. This layer can have point, line, or polygon features. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    measurement_type             Required string. Specify the mode of transportation for the analysis.

                                 Choice list: ['StraightLine', 'Driving Distance', 'Driving Time ', 'Rural Driving Distance', 'Rural Driving Time', 'Trucking Distance', 'Trucking Time', 'Walking Distance', 'Walking Time']

                                 The default is 'StraightLine'.
    -------------------------    ---------------------------------------------------------
    max_count                    Optional string. The maximum number of nearest locations to find for each feature in ``analysis_layer``. The default is the maximum cutoff allowed by the service, which is 100.

                                 Note that setting a maxCount for this parameter doesn't guarantee that many features will be found. The ``search_cutoff`` and other constraints may also reduce the number of features found.
    -------------------------    ---------------------------------------------------------
    search_cutoff                Optional float. The maximum range to search for nearest locations from each feature in the ``analysis_layer``.
                                 The units for this parameter is always minutes when ``measurement_type`` is set to a time based travel mode;
                                 otherwise the units are set in the ``search_cutoff_units`` parameter.

                                 The default is to search without bounds.
    -------------------------    ---------------------------------------------------------
    search_cutoff_units          The units of the ``search_cutoff`` parameter. This parameter is ignored when ``measurement_type`` is set to a time based travel
                                 mode because the units for ``search_cutoff`` are always minutes in those cases. If ``measurement_type`` is set to StraightLine or another distance-based travel mode, and a value for ``search_cutoff`` is specified, set the cutoff units using this parameter.

                                 Choice list: ['Kilometers', 'Meters', 'Miles', 'Feet', '']

                                 The default value is null, which causes the service to choose either miles or kilometers according to the units property of the user making the request.
    -------------------------    ---------------------------------------------------------
    time_of_day                  Optional datetime.datetime. Specify whether travel times should consider traffic conditions. To use traffic in the analysis, set ``measurement_type`` to a travel mode object whose impedance_attribute_name property is set to travel_time and assign a value to ``time_of_day``. (A travel mode with other impedance_attribute_name values don't support traffic.) The ``time_of_day`` value represents the time at which travel begins, or departs, from the origin points. The time is specified as datetime.datetime.

                                 The service supports two kinds of traffic: typical and live. Typical traffic references travel speeds that are made up of historical averages for each five-minute interval spanning a week. Live traffic retrieves speeds from a traffic feed that processes phone probe records, sensors, and other data sources to record actual travel speeds and predict speeds for the near future.

                                 The `data coverage <http://www.arcgis.com/home/webmap/viewer.html?webmap=b7a893e8e1e04311bd925ea25cb8d7c7>`_ page shows the countries Esri currently provides traffic data for.

                                 Typical Traffic:

                                 To ensure the task uses typical traffic in locations where it is available, choose a time and day of the week, and then convert the day of the week to one of the following dates from 1990:

                                 * Monday - 1/1/1990
                                 * Tuesday - 1/2/1990
                                 * Wednesday - 1/3/1990
                                 * Thursday - 1/4/1990
                                 * Friday - 1/5/1990
                                 * Saturday - 1/6/1990
                                 * Sunday - 1/7/1990
                                 Set the time and date as datetime.datetime.

                                 For example, to solve for 1:03 p.m. on Thursdays, set the time and date to 1:03 p.m., 4 January 1990; and convert to datetime eg. datetime.datetime(1990, 1, 4, 1, 3).

                                 Live Traffic:

                                 To use live traffic when and where it is available, choose a time and date and convert to datetime.

                                 Esri saves live traffic data for 12 hours and references predictive data extending 12 hours into the future. If the time and date you specify for this parameter is outside the 24-hour time window, or the travel time in the analysis continues past the predictive data window, the task falls back to typical traffic speeds.

                                 Examples:
                                 from datetime import datetime

                                 * "time_of_day": datetime(1990, 1, 4, 1, 3) # 13:03, 4 January 1990. Typical traffic on Thursdays at 1:03 p.m.
                                 * "time_of_day": datetime(1990, 1, 7, 17, 0) # 17:00, 7 January 1990. Typical traffic on Sundays at 5:00 p.m.
                                 * "time_of_day": datetime(2014, 10, 22, 8, 0) # 8:00, 22 October 2014. If the current time is between 8:00 p.m., 21 Oct. 2014 and 8:00 p.m., 22 Oct. 2014, live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.
                                 * "time_of_day": datetime(2015, 3, 18, 10, 20) # 10:20, 18 March 2015. If the current time is between 10:20 p.m., 17 Mar. 2015 and 10:20 p.m., 18 Mar. 2015, live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.

    -------------------------    ---------------------------------------------------------
    time_zone_for_time_of_day    Optional string. Specify the time zone or zones of the ``time_of_day`` parameter.

                                 Choice list: ['GeoLocal', 'UTC']

                                 ``GeoLocal``-refers to the time zone in which the origins_layer points are located.

                                 ``UTC``-refers to Coordinated Universal Time.
    -------------------------    ---------------------------------------------------------
    include_route_layers         Optional boolean. When ``include_route_layers`` is set to True, each route from the result is also saved as a route layer item.
                                 A route layer includes all the information for a particular route such as the stops assigned to the route as well
                                 as the travel directions. Creating route layers is useful if you want to share individual routes with other members in your organization.
                                 The route layers use the output feature service name provided in the ``output_name`` parameter as a prefix and the route name generated as part
                                 of the analysis is added to create a unique name for each route layer.

                                 **Caution:**

                                 Route layers cannot be created when the output is a feature collection. The task will raise an error if ``output_name`` is not
                                 specified (which indicates feature collection output) and ``include_route_layers`` is True.

                                 The maximum number of route layers that can be created is 1,000. If the result contains more than 1,000 routes
                                 and ``include_route_layers`` is True, the task will only create the output feature service.
    -------------------------    ---------------------------------------------------------
    point_barrier_layer          Optional layer. Specify one or more point features that act as temporary restrictions (in other words, barriers) when traveling on the underlying streets.

                                 A point barrier can model a fallen tree, an accident, a downed electrical line, or anything that completely blocks traffic at a specific
                                 position along the street. Travel is permitted on the street but not through the barrier.
    -------------------------    ---------------------------------------------------------
    line_barrier_layer           Optional layer. Specify one or more line features that prohibit travel anywhere the lines intersect the streets.

                                 A line barrier prohibits travel anywhere the barrier intersects the streets. For example, a parade or protest that blocks traffic across
                                 several street segments can be modeled with a line barrier.
    -------------------------    ---------------------------------------------------------
    polygon_barrier_layer        Optional layer. Specify one or more polygon features that completely restrict travel on the streets intersected by the polygons.

                                 One use of this type of barrier is to model floods covering areas of the street network and making road travel there impossible.
    -------------------------    ---------------------------------------------------------
    output_name                  Optional string. Output feature layer collection. If not provided, a feature collection is returned.
    -------------------------    ---------------------------------------------------------
    context                      Optional dict. Context contains additional settings that affect task execution. For ``find_nearest``, there are two settings.

                                 #. Extent (``extent``)-a bounding box that defines the analysis area. Only those points in the ``input_layer``
                                    that intersect the bounding box will be analyzed.

                                 #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    -------------------------    ---------------------------------------------------------
    gis                          Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------    ---------------------------------------------------------
    estimate                     Optional boolean. If True, the estimated number of credits required to run the operation will be returned.
    -------------------------    ---------------------------------------------------------
    future                       Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =========================    =========================================================

    :Returns:

    dict with the following keys:

       "nearest_layer" : layer (FeatureCollection)

       "connecting_lines_layer" : layer (FeatureCollection)

    .. code-block:: python

        #USAGE EXAMPLE: To find which regional office can be reached in the shortest drive time from esri headquarter.

        result1 = find_nearest(analysis_layer=esri_hq_lyr,
                               near_layer=regional_offices_lyr,
                               measurement_type="Driving Time",
                               output_name="find nearest office",
                               include_route_layers=True,
                               point_barrier_layer=road_closures_lyr))
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    if isinstance(measurement_type, str):
          route_service = network.RouteLayer(gis.properties.helperServices.route.url, gis=gis)
          travelmodes = route_service.retrieve_travel_modes()

          for tm in travelmodes['supportedTravelModes']:
            if tm['name'] == measurement_type:
              tm = [i for i in route_service.retrieve_travel_modes()['supportedTravelModes'] if i['name'] == measurement_type][0]
              measurement_type = tm
         
    return gis._tools.featureanalysis.find_nearest(
        analysis_layer,
        near_layer,
        measurement_type,
        max_count,
        search_cutoff,
        search_cutoff_units,
        _date_handler(time_of_day),
        time_zone_for_time_of_day,
        output_name,
        context,
        estimate=estimate,
        include_route_layers=include_route_layers,
        point_barrier_layer=point_barrier_layer,
        line_barrier_layer=line_barrier_layer,
        polygon_barrier_layer=polygon_barrier_layer,
        future=future)


def plan_routes(
        stops_layer,
        route_count,
        max_stops_per_route,
        route_start_time,
        start_layer,
        start_layer_route_id_field=None,
        return_to_start=True,
        end_layer=None,
        end_layer_route_id_field=None,
        travel_mode="Driving Time",
        stop_service_time=0,
        max_route_time=525600,
        include_route_layers=False,
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        point_barrier_layer=None,
        line_barrier_layer=None,
        polygon_barrier_layer=None,
        future=False):
    """

    .. image:: _static/images/plan_routes/plan_routes.png

    .. |balanced| image:: _static/images/plan_routes/balanced.png
    .. |partially_balanced| image:: _static/images/plan_routes/partially_balanced.png
    .. |unbalanced| image:: _static/images/plan_routes/unbalanced.png

    The ``plan_routes`` method determines how to efficiently divide tasks among a mobile workforce.

    You provide the input, which includes a set of stops and the number of vehicles available to
    visit the stops, and the tool assigns the stops to vehicles and returns routes showing how each
    vehicle can reach their assigned stops in the least amount of time.

    With ``plan_routes``, mobile workforces reach more jobsites in less time, which increases
    productivity and improves customer service. Organizations often use ``plan_routes`` to:

    * Inspect homes, restaurants, and construction sites
    * Provide repair, installation, and technical services
    * Deliver items and small packages
    * Make sales calls
    * Provide van transportation from spectators' homes to events

    The output from ``plan_routes`` includes a layer of routes showing the shortest paths to visit
    the stops; a layer of the stops assigned to routes, as well as any stops that couldn't be reached
    due to the given parameter settings; and a layer of directions containing the travel itinerary for each route.

    ============================    ==================================================================================================
    **Parameter**                   **Description**
    ----------------------------    --------------------------------------------------------------------------------------------------
    stops_layer                     Required feature layer. The points that the vehicles, drivers, or routes, should visit.
                                    The fields on the input stops are included in the output stops, so if your input
                                    layer has a field such as Name, Address, or ProductDescription, that information
                                    will be available in the results. See :ref:`Feature Input<FeatureInput>`.
    ----------------------------    --------------------------------------------------------------------------------------------------
    route_count                     Required integer. The number of vehicles that are available to visit the stops.
                                    The method supports up to 100 vehicles.

                                    The default value is 0.

                                    The method may be able to find and return a solution that uses fewer vehicles than
                                    the number you specify for this parameter. The number of vehicles returned also
                                    depends on four other parameters: the total number of stops in ``stops_layer``, the
                                    number of stops per vehicle you allow (``max_stops_per_route``), the travel time between
                                    stops, the time spent at each stop (``stop_service_time``), and any limit you set on the
                                    total route time per vehicle (``max_route_time``).
    ----------------------------    --------------------------------------------------------------------------------------------------
    max_stops_per_route             Required integer. The maximum number of stops a route, or vehicle, is allowed to visit.
                                    The largest value you can specify is 200. The default value is zero.

                                    This is one of two parameters that balance the overall workload across routes.
                                    The other is ``max_route_time``.

                                    By lowering the maximum number of stops that can be assigned to each vehicle, the vehicles
                                    are more likely to have an equal number of stops assigned to them. This helps
                                    balance workloads among drivers. The drawback, however, is that it may result in a
                                    solution that is less efficient.

                                    By increasing the stops per vehicle, the tool has more freedom to find more efficient solutions;
                                    however, the workload may be unevenly distributed among drivers and vehicles. Note that you can
                                    balance workloads by time instead of number of stops by specifying a value for the ``max_route_time`` parameter.

                                    The following examples demonstrate the effects of limiting the maximum stops per vehicle or the
                                    total time per vehicle. In all of these examples, two routes start at the same location
                                    and visit a total of six stops.

                                    +----------------------------+----------------------------------------------------------------------------------------------------------+
                                    | |balanced|                 | Balanced travel times and stops per route:                                                               |
                                    |                            |                                                                                                          |
                                    |                            | The stops are more or less uniformly spread apart, so setting ``max_stops_per_route``=3 to evenly        |
                                    |                            | distribute the workload results in routes that are roughly the same duration.                            |
                                    |                            |                                                                                                          |
                                    +----------------------------+----------------------------------------------------------------------------------------------------------+
                                    | |partially_balanced|       | Balanced stops per route but unbalanced travel times:                                                    |
                                    |                            |                                                                                                          |
                                    |                            | Five of the six stops are clustered near the starting location, but one stop is set apart                |
                                    |                            | and requires a much longer drive to be reached. Dividing the stops equally between the two               |
                                    |                            | routes (``max_stops_per_route``=3) causes unbalanced travel times.                                       |
                                    +----------------------------+----------------------------------------------------------------------------------------------------------+
                                    | |unbalanced|               | Unbalanced stops per route but balanced travel times:                                                    |
                                    |                            |                                                                                                          |
                                    |                            | The stops are in the same location as the previous graphic. By increasing the value of                   |
                                    |                            | ``max_stops_per_route`` to 4, and limiting the total travel time per vehicle (``max_route_time``),       |
                                    |                            | the travel times are balanced even though one route visits more stops.                                   |
                                    +----------------------------+----------------------------------------------------------------------------------------------------------+

    ----------------------------    --------------------------------------------------------------------------------------------------
    route_start_time                Required datetime.datetime. Specify when the vehicles or people start their routes.
                                    The time is specified as datetime.
                                    The starting time value is the same for all routes; that is, all routes start at the same time.

                                    Time zones affect what value you assign to ``route_start_time``. The time zone for the start time is
                                    based on the time zone in which the starting point is geographically located. For instance,
                                    if you have one route starting location and it is located in Pacific Standard Time (PST),
                                    the time you specify for ``route_start_time`` is in PST.

                                    There are a couple of scenarios to beware of given that starting times are based on where
                                    the starting points are located. One situation to be careful of is when you are located in
                                    one time zone but your starting locations are in another times zone. For instance, assume
                                    you are in Pacific Standard Time (UTC-8:00) and the vehicles you are routing are stationed
                                    in Mountain Standard Time (UTC-7:00). If it is currently 9:30 a.m. PST (10:30 a.m. MST)
                                    and your vehicles need to begin their routes in 30 minutes, you would set the start time
                                    to 11:00 a.m. That is, the starting locations for the routes are in the Mountain time zone,
                                    and it is currently 10:30 a.m. there, therefore, a starting time of 30 minutes from now is 11:00 a.m.
                                    Make sure you set the parameter according to the proper time zone.

                                    The other situation that requires caution is where starting locations are spread across
                                    multiple time zones. The time you set for ``route_start_time`` is specific to the time zone in
                                    which the starting location is regardless of whether there are one or more starting locations
                                    in the problem you submit. For instance, if one route starts from a point in PST and another
                                    route starts from MST, and you enter 11:00 a.m. as the start time, the route in PST will start
                                    at 11:00 a.m. PST and the route in MST will start at 11:00 a.m. MST a one-hour difference. The
                                    starting times are the same in local time, but offset in actual time, or UTC.

                                    The service automatically determines the time zones of the input starting locations (``start_layer``) for you.

                                    Examples:

                                    * datetime(2014, 10, 22, 8, 0) # 8:00, 22 October 2014. Routes will depart their
                                      starting locations at 8:00 a.m., 22 October. Any routes with starting points in Mountain
                                      Standard Time start at 8:00 a.m., 22 October 2014 MST; any routes with starting points in
                                      Pacific Standard Time start at 8:00 a.m. 22 October 2014 PST, and so on.
                                    * datetime(2015, 3, 18, 10, 20) # 10:20, 18 March 2015.
    ----------------------------    --------------------------------------------------------------------------------------------------
    start_layer                     Required feature layer. Provide the locations where the people or vehicles start their routes.
                                    You can specify one or many starting locations.

                                    If specifying one, all routes will start from the one location. If specifying many starting
                                    locations, each route needs exactly one predefined starting location, and the following criteria must be met:

                                    The number of routes (``route_count``) must equal the number of points in ``start_layer``. (However,
                                    when only one point is included in ``start_layer``, it is assumed that all routes start from
                                    the same location, and the two numbers can be different.)
                                    The starting location for each route must be identified with the ``start_layer_route_id_field``
                                    parameter. This implies that the input points in ``start_layer`` have a unique identifier.
                                    Bear in mind that if you also have many ending locations, those locations need to be
                                    predetermined as well. The predetermined start and end locations of each route are
                                    paired together by matching route ID values.
                                    See the the section of this topic entitled Starting and ending locations of
                                    routes to learn more. See :ref:`Feature Input<FeatureInput>`.
    ----------------------------    --------------------------------------------------------------------------------------------------
    start_layer_route_id_field      Optional string. Choose a field that uniquely identifies points in start_layer.
                                    This parameter is required when ``start_layer`` has more than one point; it is ignored otherwise.

                                    The ``start_layer_route_id_field`` parameter helps identify where routes begin and
                                    indicates the names of the output routes.

                                    See the the section of this topic entitled Starting and ending locations
                                    of routes to learn more.
    ----------------------------    --------------------------------------------------------------------------------------------------
    return_to_start                 Optional boolean. A True value indicates each route must end its trip at the same place where
                                    it started. The starting location is defined by the ``start_layer`` and ``start_layer_route_id_field`` parameters.

                                    The default value is True.
    ----------------------------    --------------------------------------------------------------------------------------------------
    end_layer                       Optional layer. Provide the locations where the people or vehicles end their routes.

                                    If ``end_layer`` is not specified, ``return_to_start`` must be set to True.

                                    You can specify one or many ending locations.

                                    If specifying one, all routes will end at the one location. If specifying many ending
                                    locations, each route needs exactly one predefined ending location, and the following criteria must be met:

                                    + The number of routes (``route_count``) must equal the number of points in ``end_layer``.
                                     (However, when only one point is included in ``end_layer``, it is assumed that all routes
                                     end at the same location, and the two numbers can be different.)
                                    + The ending location for each route must be identified with the ``start_layer_route_id_field``
                                     parameter. This implies that the input points in endLayer have a unique identifier.
                                     Bear in mind that if you also have many starting locations, those locations need to be
                                     predetermined as well. The predetermined start and end locations of each route are paired
                                     together by matching route ID values. See :ref:`Feature Input<FeatureInput>`.
    ----------------------------    --------------------------------------------------------------------------------------------------
    end_layer_route_id_field        Optional string. Choose a field that uniquely identifies points in ``end_layer``.
                                    This parameter is required when ``end_layer`` has more than one point; it is ignored
                                    if there is one point or if ``return_to_start`` is True.

                                    The ``end_layer_route_id_field`` parameter helps identify where routes end and indicates the names of the output routes.

                                    See the the section of this topic entitled Starting and ending locations of routes to learn more.
    ----------------------------    --------------------------------------------------------------------------------------------------
    travel_mode                     Optional string. Optional string. Specify the mode of transportation for the analysis.

                                    Choice list: ['Driving Distance', 'Driving Time', 'Rural Driving Distance', 'Rural Driving Time', 'Trucking Distance', 'Trucking Time', 'Walking Distance', 'Walking Time']
    ----------------------------    --------------------------------------------------------------------------------------------------
    stop_service_time               Optional float. Indicates how much time, in minutes, is spent at each stop.
                                    The units are minutes. All stops are assinged the same service duration from
                                    this parameter unique values for individual stops cannot be specified with this service.
    ----------------------------    --------------------------------------------------------------------------------------------------
    max_route_time                  Optional float. The amount of time you specify here limits the maximum duration of each route.
                                    The maximum route time is an accumulation of travel time and the total service time at visited
                                    stops (``stop_service_time``). This parameter is commonly used to prevent drivers from working
                                    too many hours or to balance workloads across routes or drivers.

                                    The units are 'minutes'. The default value, which is also the maximum value, is 525600 minutes, or one year.
    ----------------------------    --------------------------------------------------------------------------------------------------
    include_route_layers            Optional boolean. When ``include_route_layers`` is set to True, each route from the result is also
                                    saved as a route layer item. A route layer includes all the information for a particular route such as the stops assigned to
                                    the route as well as the travel directions. Creating route layers is useful if you want to share individual routes with other
                                    members in your organization. The route layers use the output feature service name provided in the ``output_name`` parameter as a
                                    prefix and the route name generated as part of the analysis is added to create a unique name for each route layer.
    ----------------------------    --------------------------------------------------------------------------------------------------
    output_name                     Optional string. If provided, the method will create a feature service of the results. You define
                                    the name of the service. If ``output_name`` is not supplied, the method will return a feature collection.
    ----------------------------    --------------------------------------------------------------------------------------------------
    context                         Optional string. Context contains additional settings that affect task execution. For ``plan_routes``, there are two settings:

                                    #.  Extent (``extent``) - a bounding box that defines the analysis area. Only those points in the inputLayer, start_layer, and endLayer that are within the bounding box can be visited by routes.

                                    #. Output Spatial Reference (``outSR``)-If the output is a feature service, the spatial reference will be the same as ``stops_layer``. Setting outSR for feature services has no effect.
                                       If the output is a feature collection, the features will be in the spatial reference of the outSR value or the spatial reference of ``stops_layer`` when outSR is not specified.
    ----------------------------    --------------------------------------------------------------------------------------------------
    gis                             Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    ----------------------------    --------------------------------------------------------------------------------------------------
    estimate                        Optional boolean. If True, the number of credits to run the operation will be returned.
    ----------------------------    --------------------------------------------------------------------------------------------------
    point_barrier_layer             Optional feature layer. Specify one or more point features that act as temporary restrictions (in other words, barriers) when traveling on the underlying streets.

                                    A point barrier can model a fallen tree, an accident, a downed electrical line, or anything that completely blocks traffic at a specific position along the street. Travel is permitted on the street but not through the barrier. See :ref:`Feature Input<FeatureInput>`.
    ----------------------------    --------------------------------------------------------------------------------------------------
    line_barrier_layer              Optional feature layer. Specify one or more line features that prohibit travel anywhere the lines intersect the streets.

                                    A line barrier prohibits travel anywhere the barrier intersects the streets. For example, a parade or protest that blocks traffic across several street segments can be modeled with a line barrier. See :ref:`Feature Input<FeatureInput>`.
    ----------------------------    --------------------------------------------------------------------------------------------------
    polygon_barrier_layer           Optional feature layer. Specify one or more polygon features that completely restrict travel on the streets intersected by the polygons.

                                    One use of this type of barrier is to model floods covering areas of the street network and making road travel there impossible. See :ref:`Feature Input<FeatureInput>`.
    ----------------------------    --------------------------------------------------------------------------------------------------
    future                          Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ============================    ==================================================================================================

    :returns: feature layer Item if ``output_name`` is specified, else dict with the following keys:


        "routes_layer" : layer (FeatureCollection)

        "assigned_stops_layer" : layer (FeatureCollection)

        "unassigned_stops_layer" : layer (FeatureCollection)

    .. code-block:: python

        # USAGE EXAMPLE: To plan routes to provide cab from employee residence to office.
        route = plan_routes(stops_layer=employee_residence,
                            route_count=4,
                            max_stops_per_route=4,
                            route_start_time=datetime(2019, 6, 20, 6, 0),
                            start_layer=office_location,
                            start_layer_route_id_field='n_office',
                            return_to_start=False,
                            end_layer=office_location,
                            end_layer_route_id_field='n_office',
                            travel_mode='Driving Time',
                            stop_service_time=5,
                            include_route_layers=False,
                            output_name='plan route for employees')

    """
    gis = _arcgis.env.active_gis if gis is None else gis
    if isinstance(travel_mode, str):
        route_service = network.RouteLayer(gis.properties.helperServices.route.url, gis=gis)
        travelmodes = route_service.retrieve_travel_modes()
        for tm in travelmodes['supportedTravelModes']:
            if tm['name'] == travel_mode:
              tm = [i for i in route_service.retrieve_travel_modes()['supportedTravelModes'] if i['name'] == travel_mode][0]
              travel_mode = tm
    return gis._tools.featureanalysis.plan_routes(
        stops_layer,
        route_count,
        max_stops_per_route,
        _date_handler(route_start_time),
        start_layer,
        start_layer_route_id_field,
        return_to_start,
        end_layer,
        end_layer_route_id_field,
        travel_mode,
        stop_service_time,
        max_route_time,
        include_route_layers,
        output_name,
        context,
        estimate=estimate,
        point_barrier_layer=point_barrier_layer,
        line_barrier_layer=line_barrier_layer,
        polygon_barrier_layer=polygon_barrier_layer, future=future)



import os, json, tempfile
from pathlib import Path
from ._codetemplate import image_classifier_prf
from ._arcgis_model import _raise_fastai_import_error
from functools import partial
from ._arcgis_model import ArcGISModel

try:
    from fastai.basic_train import Learner
    from ._arcgis_model import SaveModelCallback
    from ._unet_utils import is_no_color
    import torch
    from torch import nn
    import torch.nn.functional as F
    from torchvision import models
    from ._unet_utils import LabelCallback
    from ._arcgis_model import _EmptyData
    from fastai.vision import to_device
    from ._psp_utils import PSPNet, _pspnet_unet
    from fastai.vision.models import unet
    import numpy as np
    from fastai.callbacks import EarlyStoppingCallback
    from fastai.torch_core import split_model_idx
    from fastai.vision import flatten_model
    HAS_FASTAI = True
except Exception as e:
    HAS_FASTAI = False

def _pspnet_learner(data,  backbone, chip_size=224, pyramid_sizes=(1, 2, 3, 6), pretrained=True, **kwargs):
    "Build psp_net learner from `data` and `arch`."
    model = to_device(PSPNet(data.c, backbone, chip_size, pyramid_sizes, pretrained), data.device)
    learn = Learner(data, model, **kwargs)
    return learn

def _pspnet_learner_with_unet(data,  backbone, chip_size=224, pyramid_sizes=(1, 2, 3, 6), pretrained=True, **kwargs):
    "Build psp_net learner from `data` and `arch`."
    model = unet.DynamicUnet(encoder=_pspnet_unet(data.c, backbone, chip_size, pyramid_sizes, pretrained), n_classes=data.c, last_cross=False)
    learn = Learner(data, model, **kwargs)
    return learn


class PSPNetClassifier(ArcGISModel):

    """
    Model architecture from https://arxiv.org/abs/1612.01105.
    Creates a PSPNet Image Segmentation/ Pixel Classification model. 

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the `PSPNetClassifier`, which
                            is `resnet50` by default. It supports the ResNet,
                            DenseNet, and VGG families.
    ---------------------   -------------------------------------------
    use_unet                Optional Bool. Specify whether to use Unet-Decoder or not,
                            Default True.                          
    ---------------------   -------------------------------------------
    pyramid_sizes           Optional List. The sizes at which the feature map is pooled at.
                            Currently set to the best set reported in the paper,
                            i.e, (1, 2, 3, 6)
    ---------------------   -------------------------------------------
    pretrained              Optional Bool. If True, use the pretrained backbone                                                                                 
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained PSPNet model is
                            saved.
    =====================   ===========================================

    :returns: `PSPNetClassifier` Object
    """

    def __init__(self, data, backbone=None, use_unet=True, pyramid_sizes=[1, 2, 3, 6], pretrained_path=None):
        # Set default backbone to be 'resnet50'
        if backbone is None: 
            backbone = models.resnet50
      
        super().__init__(data, backbone)     

        # Check if a backbone provided is compatible, use resnet50 as default
        if not self._check_backbone_support(backbone):
            raise Exception (f"Enter only compatible backbones from {', '.join(self.supported_backbones)}")              

        self._code = image_classifier_prf
        self.pyramid_sizes = pyramid_sizes
        self._use_unet = use_unet
        if use_unet:
            self.learn = _pspnet_learner_with_unet(data, backbone=self._backbone, chip_size=self._data.chip_size, pyramid_sizes=pyramid_sizes, pretrained=True, metrics=self.accuracy)
        else:
            self.learn = _pspnet_learner(data, backbone=self._backbone, chip_size=self._data.chip_size, pyramid_sizes=pyramid_sizes, pretrained=True, metrics=self.accuracy)
            self.learn.loss_func = self._psp_loss
        self.learn.callbacks.append(LabelCallback(self.learn))  #appending label callback 

        if pretrained_path is not None:
            self.load(pretrained_path)
        
        self.freeze()

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    # Return a list of supported backbones names
    @property
    def supported_backbones(self):
        return [*self._resnet_family, *self._densenet_family, *self._vgg_family]

    @classmethod
    def from_model(cls, emd_path, data=None):
        emd_path = Path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)
            
        model_file = Path(emd['ModelFile'])
        
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file
            
        model_params = emd['ModelParameters']

        try:
            class_mapping = {i['Value'] : i['Name'] for i in emd['Classes']}
            color_mapping = {i['Value'] : i['Color'] for i in emd['Classes']}
        except KeyError:
            class_mapping = {i['ClassValue'] : i['ClassName'] for i in emd['Classes']} 
            color_mapping = {i['ClassValue'] : i['Color'] for i in emd['Classes']}                

        
        if data is None:
            empty_data = _EmptyData(path=tempfile.TemporaryDirectory().name, loss_func=None, c=len(class_mapping) + 1, chip_size=emd['ImageHeight'])
            empty_data.class_mapping = class_mapping
            empty_data.color_mapping = color_mapping
            return cls(empty_data, **model_params, pretrained_path=str(model_file))
        else:
            return cls(data, **model_params, pretrained_path=str(model_file)) 


    def accuracy(self, input, target, void_code=0, class_mapping=None): 
        if self.learn.model.training: # while training
            input = input[0]

        target = target.squeeze(1)
        mask = target != void_code
        return (input.argmax(dim=1)[mask] == target[mask]).float().mean()

    def _psp_loss(self, outputs, targets):
        targets = targets.squeeze().detach()
        criterion = nn.CrossEntropyLoss().cuda()

        if self.learn.model.training: # returns a tuple of aux_logits and main_logits while training
            out = outputs[0]
            aux = outputs[1]
        else: # validation
            out = outputs

        main_loss = criterion(out, targets)

        if self.learn.model.training:
            aux_loss = criterion(aux, targets)
            total_loss = main_loss + 0.4 * aux_loss
            return total_loss
        else:
            return main_loss

    def freeze(self):
        "Freezes the pretrained backbone."
        for idx, i in enumerate(flatten_model(self.learn.model)):
            if hasattr(i, 'dilation'):
                dilation = i.dilation
                dilation = dilation[0] if isinstance(dilation, tuple) else dilation
                if dilation > 1:
                    break        
            for p in i.parameters():
                p.requires_grad = False

        self.learn.layer_groups = split_model_idx(self.learn.model, [idx])  ## Could also call self.learn.freeze after this line because layer groups are now present.      
  
    def unfreeze(self):
        for _, param in self.learn.model.named_parameters():
            param.requires_grad = True
        
    def _create_emd(self, path):
        import random
        super()._create_emd(path)
        
        self._emd_template["ModelParameters"]["pyramid_sizes"] = self.pyramid_sizes
        self._emd_template["ModelParameters"]["use_unet"] = self._use_unet
        self._emd_template["Framework"] = "arcgis.learn.models._inferencing"
        self._emd_template["ModelConfiguration"] = "_psp"
        self._emd_template["InferenceFunction"] = "ArcGISImageClassifier.py"
        self._emd_template["ExtractBands"] = [0, 1, 2]

        self._emd_template['Classes'] = []
        class_data = {}
        for i, class_name in enumerate(self._data.classes[1:]):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)] if is_no_color(self._data.color_mapping) else \
            self._data.color_mapping[inverse_class_mapping[class_name]]
            class_data["Color"] = color
            self._emd_template['Classes'].append(class_data.copy())

        json.dump(self._emd_template, open(path.with_suffix('.emd'), 'w'), indent=4)
        return path.stem

    def show_results(self, rows=5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        """
        if rows > self._data.batch_size:
            rows = self._data.batch_size
        self.learn.show_results(rows=rows, **kwargs)   

    @property
    def _model_metrics(self):
        return {'accuracy': self._get_model_metrics()}

    def _get_model_metrics(self, **kwargs):
        checkpoint = kwargs.get('checkpoint', True)
        model_accuracy = self.learn.recorder.metrics[-1][0]
        if checkpoint:
            model_accuracy = np.min(self.learn.recorder.metrics)             
        return float(model_accuracy)


from ._arcgis_model import ArcGISModel
import tempfile
from pathlib import Path
import json
from ._codetemplate import code
import logging
logger = logging.getLogger() 
import os, csv
import warnings
from warnings import warn
from . import _tracker_util

HAS_OPENCV = True
HAS_FASTAI = True
HAS_ARCPY = True

try:
    import torch
    import numpy as np
    from fastprogress import progress_bar
    from fastai.vision.learner import cnn_learner
    from fastai.callbacks.hooks import model_sizes
    from fastai.vision.learner import create_body
    from fastai.vision.data import ImageDataBunch
    from fastai.vision import ImageList
    from fastai.vision import imagenet_stats, normalize
    from fastai.vision.image import open_image, bb2hw, image2np, Image, pil2tensor
    from torchvision.models import resnet34
    from torchvision.models import mobilenet_v2
    from torchvision import models
    from ._ssd_utils import SSDHead, BCE_Loss, FocalLoss, one_hot_embedding, nms
    from ._ssd_utils import SSDObjectCategoryList, compute_class_AP, SSDHeadv2, kmeans, avg_iou
    from .._data import prepare_data
    from fastai.callbacks import EarlyStoppingCallback
    from ._arcgis_model import SaveModelCallback, _set_multigpu_callback
    from ._unet_utils import is_no_color
    from torch.nn import Module as NnModule
    import PIL
    from .._image_utils import _get_image_chips, _get_transformed_predictions, _draw_predictions, _exclude_detection
    from .._video_utils import VideoUtils
except Exception as e:
    class NnModule():
        pass
    HAS_FASTAI = False

try:
    import cv2
except Exception:
    HAS_OPENCV = False

try:
    import arcpy
except Exception:
    HAS_ARCPY = False


def _mobilenet_split(m:NnModule): return m[0][0][0], m[1]


class SingleShotDetector(ArcGISModel):

    """
    Creates a Single Shot Detector with the specified grid sizes, zoom scales
    and aspect  ratios. Based on Fast.ai MOOC Version2 Lesson 9.
    
    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    grids                   Required list. Grid sizes used for creating anchor
                            boxes.
    ---------------------   -------------------------------------------
    zooms                   Optional list. Zooms of anchor boxes.
    ---------------------   -------------------------------------------
    ratios                  Optional list of tuples. Aspect ratios of anchor
                            boxes.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the `SingleShotDetector`, which
                            is `resnet34` by default.
    ---------------------   -------------------------------------------
    dropout                 Optional float. Dropout propbability. Increase it to
                            reduce overfitting.
    ---------------------   -------------------------------------------
    bias                    Optional float. Bias for SSD head.
    ---------------------   -------------------------------------------
    focal_loss              Optional boolean. Uses Focal Loss if True.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    ---------------------   -------------------------------------------
    location_loss_factor    Optional float. Sets the weight of the bounding box
                            loss. This should be strictly between 0 and 1. This 
                            is default `None` which gives equal weight to both 
                            location and classification loss. This factor
                            adjusts the focus of model on the location of 
                            bounding box.
    ---------------------   -------------------------------------------
    ssd_version             Optional int within [1,2]. Use version=1 for arcgis v1.6.2 or earlier
    =====================   ===========================================
    
    :returns: `SingleShotDetector` Object
    """

    def __init__(self, data, grids=None, zooms=[1.], ratios=[[1., 1.]],
                 backbone=None, drop=0.3, bias=-4., focal_loss=False, pretrained_path=None, location_loss_factor=None, ssd_version=2):

        super().__init__(data, backbone)

        # assert (location_loss_factor is not None) or ((location_loss_factor > 0) and (location_loss_factor < 1)),
        if not ssd_version in [1, 2]:
            raise Exception("ssd_version can be only [1,2]")

        if location_loss_factor is not None:
            if not ((location_loss_factor > 0) and (location_loss_factor < 1)):
                raise Exception('`location_loss_factor` should be greater than 0 and less than 1')
        self.location_loss_factor = location_loss_factor

        self._code = code
        self.ssd_version = ssd_version

        if backbone is None:
            self._backbone = models.resnet34
            backbone_name = 'res'
        elif type(backbone) is str:
            self._backbone = getattr(models, backbone)
            backbone_name = backbone[:3]
        else:
            self._backbone = backbone
            backbone_name = 'custom'

        if not self._check_backbone_support(self._backbone):
            raise Exception (f"Enter only compatible backbones from {', '.join(self.supported_backbones)}")

        backbone_cut = None
        backbone_split = None

        if self._backbone == models.mobilenet_v2:
            backbone_cut = -1
            backbone_split = _mobilenet_split

        if ssd_version == 1:
            if grids == None:
                grids =[4,2,1]
                
            self._create_anchors(grids, zooms, ratios)

            feature_sizes = model_sizes(create_body(self._backbone), size=(data.chip_size, data.chip_size))
            num_features = feature_sizes[-1][-1]
            num_channels = feature_sizes[-1][1]

            ssd_head = SSDHead(grids, self._anchors_per_cell, data.c, num_features=num_features, drop=drop, bias=bias, num_channels=num_channels)
        elif ssd_version == 2:

            # find bounding boxes height and width
        
            if grids is None:
                logger.info("Computing optimal grid size...")
                hw = data.height_width
                hw = np.array(hw)
                
                # find most suitable centroids for dataset
                centroid = kmeans(hw , 1) 
                avg = avg_iou(hw, centroid)

                for num_anchor in range(2, 5):
                    new_centroid = kmeans(hw, num_anchor)
                    new_avg = avg_iou(hw, new_centroid)
                    if (new_avg - avg) < 0.05:
                       break
                    avg = new_avg
                    centroid = new_centroid.copy()

                # find grid size

                grids = list(map(int, map(round, data.chip_size/np.sort(np.max(centroid, axis=1)))))
                if grids[-1] == 0:
                    grids[-1] = 1

                grids = list(set(grids))
                grids.sort(reverse = True)
            
            self._create_anchors(grids, zooms, ratios)

            feature_sizes = model_sizes(create_body(self._backbone, cut=backbone_cut), size=(data.chip_size, data.chip_size))
            num_features = feature_sizes[-1][-1]
            num_channels = feature_sizes[-1][1] 

            if grids[0] > 8 and abs(num_features - grids[0]) > 4 and backbone_name == 'res':
                num_features = feature_sizes[-2][-1]
                num_channels = feature_sizes[-2][1]
                backbone_cut = -3
                ssd_head = SSDHeadv2(grids, self._anchors_per_cell, data.c, num_features=num_features, drop=drop, bias=bias, num_channels=num_channels)
            else:
                ssd_head = SSDHeadv2(grids, self._anchors_per_cell, data.c, num_features=num_features, drop=drop, bias=bias, num_channels=num_channels)
        else:
            raise Exception('SSDVersion can only be 1 or 2')

        self.learn = cnn_learner(data=data, base_arch=self._backbone, cut=backbone_cut, split_on=backbone_split, custom_head=ssd_head)
        self.learn.model = self.learn.model.to(self._device)

        if focal_loss:
            self._loss_f = FocalLoss(data.c)
        else:
            self._loss_f = BCE_Loss(data.c)
        self.learn.loss_func = self._ssd_loss

        _set_multigpu_callback(self)
        if pretrained_path is not None:
            self.load(pretrained_path)        

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    @property
    def supported_backbones(self):
        return [*self._resnet_family, *self._densenet_family, *self._vgg_family, models.mobilenet_v2.__name__]

    @classmethod
    def from_model(cls, emd_path, data=None):

        """
        Creates a Single Shot Detector from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        =====================   ===========================================
        
        :returns: `SingleShotDetector` Object
        """
        return cls.from_emd(data, emd_path)

    @classmethod
    def from_emd(cls, data, emd_path):

        """
        Creates a Single Shot Detector from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        =====================   ===========================================
        
        :returns: `SingleShotDetector` Object
        """
        emd_path = Path(emd_path)
        emd = json.load(open(emd_path))
        model_file = Path(emd['ModelFile'])
        backbone = emd.get('backbone', 'resnet34')
        ssd_version = int(emd.get('SSDVersion', 1))
        chip_size = emd["ImageWidth"]

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        class_mapping = {i['Value']: i['Name'] for i in emd['Classes']}
        
        resize_to = emd.get('resize_to')
        if isinstance(resize_to, list):
            resize_to = (resize_to[0], resize_to[1])

        data_passed = True
        # Create an image databunch for when loading the model using emd (without training data)
        if data is None:
            data_passed = False
            train_tfms = []
            val_tfms = []
            ds_tfms = (train_tfms, val_tfms)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                
                sd = ImageList([], path=tempfile.TemporaryDirectory().name).split_by_idx([])
                tempdata = sd.label_const(0, label_cls=SSDObjectCategoryList, classes=list(class_mapping.values())).transform(ds_tfms).databunch().normalize(imagenet_stats)
                tempdata.chip_size = chip_size
                tempdata.class_mapping = class_mapping
                tempdata.classes = ['background'] + list(class_mapping.values())
                data = tempdata
                data.c += 1 # Add 1 for background class

        data.resize_to = resize_to
        ssd = cls(data, emd['Grids'], emd['Zooms'], emd['Ratios'], pretrained_path=str(model_file), backbone=backbone, ssd_version=ssd_version)

        if not data_passed:
            ssd.learn.data.single_ds.classes = ssd._data.classes
            ssd.learn.data.single_ds.y.classes = ssd._data.classes
        
        return ssd

    def _create_anchors(self, anc_grids, anc_zooms, anc_ratios):

        self.grids = anc_grids
        self.zooms = anc_zooms
        self.ratios =  anc_ratios

        anchor_scales = [(anz*i, anz*j) for anz in anc_zooms for (i,j) in anc_ratios]

        self._anchors_per_cell = len(anchor_scales)

        anc_offsets = [1/(o*2) for o in anc_grids]

        anc_x = np.concatenate([np.repeat(np.linspace(ao, 1-ao, ag), ag)
                                for ao,ag in zip(anc_offsets,anc_grids)])
        anc_y = np.concatenate([np.tile(np.linspace(ao, 1-ao, ag), ag)
                                for ao,ag in zip(anc_offsets,anc_grids)])
        anc_ctrs = np.repeat(np.stack([anc_x,anc_y], axis=1), self._anchors_per_cell, axis=0)

        anc_sizes  =   np.concatenate([np.array([[o/ag,p/ag] for i in range(ag*ag) for o,p in anchor_scales])
                       for ag in anc_grids])

        self._grid_sizes = torch.Tensor(np.concatenate([np.array([ 1/ag  for i in range(ag*ag) for o,p in anchor_scales])
                       for ag in anc_grids])).unsqueeze(1).to(self._device)

        self._anchors = torch.Tensor(np.concatenate([anc_ctrs, anc_sizes], axis=1)).float().to(self._device)

        self._anchor_cnr = self._hw2corners(self._anchors[:,:2], self._anchors[:,2:])

    def _hw2corners(self, ctr, hw):
        return torch.cat([ctr-hw/2, ctr+hw/2], dim=1)

    def _get_y(self, bbox, clas):
        try:
            bbox = bbox.view(-1, 4)  # /sz
        except Exception:
            bbox = torch.zeros(size=[0, 4])
        bb_keep = ((bbox[:,2]-bbox[:,0])>0).nonzero()[:,0]
        return bbox[bb_keep],clas[bb_keep]

    def _actn_to_bb(self, actn, anchors, grid_sizes):
        actn_bbs = torch.tanh(actn)
        actn_centers = (actn_bbs[:,:2]/2 * grid_sizes) + anchors[:,:2]
        actn_hw = (actn_bbs[:,2:]/2+1) * anchors[:,2:]
        return self._hw2corners(actn_centers, actn_hw)

    def _map_to_ground_truth(self, overlaps, print_it=False):
        prior_overlap, prior_idx = overlaps.max(1)
        if print_it: print(prior_overlap)
        gt_overlap, gt_idx = overlaps.max(0)
        gt_overlap[prior_idx] = 1.99
        for i,o in enumerate(prior_idx): gt_idx[o] = i
        return gt_overlap, gt_idx

    def _ssd_1_loss(self, b_c, b_bb, bbox, clas, print_it=False):
        bbox,clas = self._get_y(bbox,clas)
        bbox = self._normalize_bbox(bbox)

        a_ic = self._actn_to_bb(b_bb, self._anchors, self._grid_sizes)
        overlaps = self._jaccard(bbox.data, self._anchor_cnr.data)
        try:
            gt_overlap,gt_idx = self._map_to_ground_truth(overlaps,print_it)
        except Exception as e:
            return torch.tensor(0., requires_grad=True).to(self._device), torch.tensor(0., requires_grad=True).to(self._device)
        gt_clas = clas[gt_idx]
        pos = gt_overlap > 0.4
        pos_idx = torch.nonzero(pos)[:,0]
        gt_clas[1-pos] = 0
        gt_bbox = bbox[gt_idx]
        loc_loss = ((a_ic[pos_idx] - gt_bbox[pos_idx]).abs()).mean()
        clas_loss  = self._loss_f(b_c, gt_clas)
        return loc_loss, clas_loss

    def _ssd_loss(self, pred, targ1, targ2, print_it=False):
        lcs, lls = 0., 0.
        for b_c,b_bb,bbox,clas in zip(*pred, targ1, targ2):
            loc_loss, clas_loss = self._ssd_1_loss(b_c, b_bb,bbox.to(self._device), clas.to(self._device), print_it)
            lls += loc_loss
            lcs += clas_loss
        if print_it: print('loc: {lls}, clas: {lcs}'.format(lls=lls, lcs=lcs))
        if self.location_loss_factor is None:
            return lls + lcs
        else:
            return self.location_loss_factor * lls + (1 - self.location_loss_factor) * lcs

    def _intersect(self,box_a, box_b):
        max_xy = torch.min(box_a[:, None, 2:], box_b[None, :, 2:])
        min_xy = torch.max(box_a[:, None, :2], box_b[None, :, :2])
        inter = torch.clamp((max_xy - min_xy), min=0)
        return inter[:, :, 0] * inter[:, :, 1]

    def _box_sz(self, b):
        return (b[:, 2]-b[:, 0]) * (b[:, 3]-b[:, 1])

    def _jaccard(self, box_a, box_b):
        inter = self._intersect(box_a, box_b)
        union = self._box_sz(box_a).unsqueeze(1) + self._box_sz(box_b).unsqueeze(0) - inter
        return inter / union

    def _normalize_bbox(self, bbox):
        return (bbox+1.)/2.

    @property
    def _model_metrics(self):
        return {'average_precision_score': self.average_precision_score(show_progress=False)}

    def _create_emd(self, path):
        import random
        super()._create_emd(path)

        self._emd_template["Framework"] = "arcgis.learn.models._inferencing"
        self._emd_template["InferenceFunction"] = "ArcGISObjectDetector.py"
        self._emd_template["ModelConfiguration"] = "_DynamicSSD"
        self._emd_template["ModelType"] = "ObjectDetection"
        self._emd_template["ExtractBands"] = [0, 1, 2]
        self._emd_template['backbone'] = self._backbone.__name__
        self._emd_template['Grids'] = self.grids
        self._emd_template['Zooms'] = self.zooms
        self._emd_template['Ratios'] = self.ratios
        self._emd_template['SSDVersion'] = self.ssd_version
        self._emd_template['Classes'] = []

        class_data = {}
        for i, class_name in enumerate(self._data.classes[1:]): # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            self._emd_template['Classes'].append(class_data.copy())

        json.dump(self._emd_template, open(path.with_suffix('.emd'), 'w'), indent=4)

        return path.stem

    def _create_tfonnx_emd(self, saved_path, batch_size):
        import random
        super()._create_emd(saved_path)

        self._emd_template["Framework"] = "arcgis.learn.models._inferencing"
        self._emd_template["InferenceFunction"] = "ArcGISObjectDetector.py"
        self._emd_template["ModelConfiguration"] = "_SSDTensorflow"
        self._emd_template["ModelType"] = "ObjectDetection"
        self._emd_template["ExtractBands"] = [0, 1, 2]
        self._emd_template['backbone'] = self._backbone.__name__
        self._emd_template['Grids'] = self.grids
        self._emd_template['Zooms'] = self.zooms
        self._emd_template['Ratios'] = self.ratios
        self._emd_template['SSDVersion'] = self.ssd_version
        self._emd_template['Classes'] = []
        self._emd_template['BatchSize'] = batch_size

        class_data = {}
        for i, class_name in enumerate(self._data.classes[1:]): # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            self._emd_template['Classes'].append(class_data.copy())

        json.dump(self._emd_template, open(saved_path.with_suffix('.emd'), 'w'), indent=4)

        return saved_path.stem
    
    def show_results(self, rows=5, thresh=0.5, nms_overlap=0.1):

        """
        Displays the results of a trained model on a part of the validation set.
        """ 
        if rows > self._data.batch_size:
            rows = self._data.batch_size      
        self.learn.show_results(rows=rows, thresh=thresh, nms_overlap=nms_overlap, ssd=self)

    def predict_video(
        self,
        input_video_path,
        metadata_file,
        threshold=0.5,
        nms_overlap=0.1,
        track=False,
        visualize=False,
        output_file_path=None,
        multiplex=False,
        multiplex_file_path=None,
        tracker_options={
            'assignment_iou_thrd': 0.3,
            'vanish_frames': 40,
            'detect_frames': 10
        },
        visual_options={
            'show_scores': True,
            'show_labels': True,
            'thickness': 2,
            'fontface': 0,
            'color': (255, 255, 255)
        }
    ):

        """
        Runs prediction on a video and appends the output VMTI predictions in the metadata file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        input_video_path        Required. Path to the video file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        metadata_file           Required. Path to the metadata csv file where
                                the predictions will be saved in VMTI format.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        track                   Optional bool. Set this parameter as True to
                                enable object tracking. 
        ---------------------   -------------------------------------------
        visualize               Optional boolean. If True a video is saved
                                with prediction results.
        ---------------------   -------------------------------------------
        output_file_path        Optional path. Path of the final video to be saved.
                                If not supplied, video will be saved at path input_video_path
                                appended with _prediction.
        ---------------------   -------------------------------------------
        multiplex               Optional boolean. Runs Multiplex using the VMTI detections.
        ---------------------   -------------------------------------------
        multiplex_file_path     Optional path. Path of the multiplexed video to be saved.
                                By default a new file with _multiplex.MOV extension is saved
                                in the same folder.
        ---------------------   -------------------------------------------
        tracking_options        Optional dictionary. Set different parameters for
                                object tracking. assignment_iou_thrd parameter is used
                                to assign threshold for assignment of trackers, 
                                vanish_frames is the number of frames the object should
                                be absent to consider it as vanished, detect_frames 
                                is the number of frames an object should be detected
                                to track it.
        ---------------------   -------------------------------------------
        visual_options          Optional dictionary. Set different parameters for
                                visualization.
                                show_scores boolean, to view scores on predictions,
                                show_labels boolean, to view labels on predictions,
                                thickness integer, to set the thickness level of box,
                                fontface integer, fontface value from opencv values,
                                color tuple (B, G, R), tuple containing values between
                                0-255.
        =====================   ===========================================
        
        """

        VideoUtils.predict_video(
            self,
            input_video_path,
            metadata_file,
            threshold,
            nms_overlap,
            track, visualize,
            output_file_path,
            multiplex,
            multiplex_file_path,
            tracker_options,
            visual_options
        )

    def predict(
        self,
        image_path,
        threshold=0.5,
        nms_overlap=0.1,
        return_scores=False,
        visualize=False
    ):

        """
        Runs prediction on a video and appends the output VMTI predictions in the metadata file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        image_path              Required. Path to the image file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        return_scores           Optional boolean. Will return the probability
                                scores of the bounding box predictions if True.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. Displays the image with
                                predicted bounding boxes if True.
        =====================   ===========================================
        
        :returns: 'List' of xmin, ymin, width, height of predicted bounding boxes on the given image
        """
        if not HAS_OPENCV:
            raise Exception("This function requires opencv 4.0.1.24. Install it using pip install opencv-python==4.0.1.24")

        if isinstance(image_path, str):
            image = cv2.imread(image_path)
        else:
            image = image_path

        orig_height, orig_width, _ = image.shape
        orig_frame = image.copy()

        if self._data.resize_to is not None:
            if isinstance(self._data.resize_to, tuple):
                image = cv2.resize(image, self._data.resize_to)
            else:
                image = cv2.resize(image, (self._data.resize_to, self._data.resize_to))

        height, width, _ = image.shape

        if self._data.chip_size is not None:
            chips = _get_image_chips(image, self._data.chip_size)
        else:
            chips = [{'width': width, 'height': height, 'xmin': 0, 'ymin': 0, 'chip': image, 'predictions': []}]

        valid_tfms = self._data.valid_ds.tfms
        self._data.valid_ds.tfms = []

        for chip in chips:
            frame = Image(pil2tensor(PIL.Image.fromarray(cv2.cvtColor(chip['chip'], cv2.COLOR_BGR2RGB)), dtype=np.float32).div_(255))
            bbox = self.learn.predict(frame, thresh=threshold, nms_overlap=nms_overlap, ret_scores=True, ssd=self)[0]
            if bbox:
                scores = bbox.scores
                bboxes, lbls = bbox._compute_boxes()
                bboxes.add_(1).mul_(torch.tensor([chip['height'] / 2, chip['width'] / 2, chip['height'] / 2, chip['width'] / 2])).long()
                for index, bbox in enumerate(bboxes):
                    if lbls is not None:
                        label = lbls[index]
                    else:
                        label = 'Default'

                    data = bb2hw(bbox)
                    if not _exclude_detection((data[0], data[1], data[2], data[3]), chip['width'], chip['height']):
                        chip['predictions'].append({
                            'xmin': data[0],
                            'ymin': data[1],
                            'width': data[2],
                            'height': data[3],
                            'score': float(scores[index]),
                            'label': label
                        })

        self._data.valid_ds.tfms = valid_tfms

        predictions, labels, scores = _get_transformed_predictions(chips)

        y_ratio = orig_height/height
        x_ratio = orig_width/width

        for index, prediction in enumerate(predictions):
            prediction[0] = prediction[0]*x_ratio
            prediction[1] = prediction[1]*y_ratio
            prediction[2] = prediction[2]*x_ratio
            prediction[3] = prediction[3]*y_ratio

            # Clip xmin
            if prediction[0] < 0: 
                prediction[2] = prediction[2] + prediction[0]
                prediction[0] = 1

            # Clip width when xmax greater than original width
            if prediction[0] + prediction[2] > orig_width:
                prediction[2] = (prediction[0] + prediction[2]) - orig_width

            # Clip ymin
            if prediction[1] < 0:
                prediction[3] = prediction[3] + prediction[1]
                prediction[1] = 1

            # Clip height when ymax greater than original height
            if prediction[1] + prediction[3] > orig_height:
                prediction[3] = (prediction[1] + prediction[3]) - orig_height

            predictions[index] = [
                prediction[0],
                prediction[1],
                prediction[2],
                prediction[3]
            ]      

        if visualize:
            image = _draw_predictions(orig_frame, predictions, labels)
            import matplotlib.pyplot as plt
            plt.xticks([])
            plt.yticks([])
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            plt.imshow(PIL.Image.fromarray(image))

        if return_scores:
            return predictions, labels, scores
        else:
            return predictions, labels

    def average_precision_score(self, detect_thresh=0.2, iou_thresh=0.1, mean=False, show_progress=True):

        """
        Computes average precision on the validation set for each class.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probabilty above which
                                a detection will be considered for computing
                                average precision.
        ---------------------   -------------------------------------------
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth labels, above
                                which a predicted bounding box will be
                                considered a true positive.
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.                        
        =====================   ===========================================
        
        :returns: `dict` if mean is False otherwise `float`
        """        
        aps = compute_class_AP(self, self._data.valid_dl, self._data.c - 1, show_progress, detect_thresh=detect_thresh, iou_thresh=iou_thresh)
        if mean:
            import statistics
            return statistics.mean(aps)
        else:
            return dict(zip(self._data.classes[1:], aps))


"""
These functions help you identify, quantify, and visualize spatial patterns in your data.

calculate_density takes known quantities of some phenomenon and spreads these quantities across the map.
find_hot_spots identifies statistically significant clustering in the spatial pattern of your data.
interpolate_points predicts values at new locations based on measurements found in a collection of points.
"""

import arcgis as _arcgis


def calculate_density(
        input_layer,
        field=None,
        cell_size=None,
        cell_size_units="Meters",
        radius=None,
        radius_units=None,
        bounding_polygon_layer=None,
        area_units=None,
        classification_type="EqualInterval",
        num_classes=10,
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        future=False):
    """
    .. image:: _static/images/cal_density_standard/calculate_density.png

    The calculate_density function creates a density map from point or line features by spreading known quantities of
    some phenomenon (represented as attributes of the points or lines) across the map. The result is a layer of areas
    classified from least dense to most dense.

    For point input, each point should represent the location of some event	or incident, and the result layer represents
    a count of the incident per unit area. A higher density value in a new location means that there are more points near
    that location. In many cases, the result layer can be interpreted as a risk surface for future events. For example,
    if the input points represent locations of lightning strikes, the result layer can be interpreted as a risk surface
    for future lightning strikes.

    For line input, the line density surface represents the total amount of line that is near each location. The units of
    the calculated density values are the length of line per unit area. For example, if the lines represent rivers, the
    result layer will represent the total length of rivers that are within the search radius. This result can be used to
    identify areas that are hospitable to grazing animals.

    =========================    =========================================================
    **Argument**                 **Description**
    -------------------------    ---------------------------------------------------------
    input_layer                  Required layer. The point or line features from which to calculate density. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    field                        Optional string. A numeric field name specifying the number of incidents at each location.  For example, if you have points that represent cities, you can use a field representing the population of the city as the count field, and the resulting population density layer will calculate larger population densities near cities with larger populations. If not specified, each location will be assumed to represent a single count.
    -------------------------    ---------------------------------------------------------
    cell_size                    Optional float. This value is used to create a mesh of points where density values are calculated. The default is approximately 1/1000th of the smaller of the width and height of the analysis extent as defined in the context parameter. The smaller the value, the smoother the polygon boundaries will be. Conversely, with larger values, the polygon boundaries will be more coarse and jagged.
    -------------------------    ---------------------------------------------------------
    cell_size_units              Optional string. The units of the cell_size value.
                                 Choice list: ['Miles', 'Feet', 'Kilometers',  'Meters']
    -------------------------    ---------------------------------------------------------
    radius                       Optional float. A distance specifying how far to search to find point or line features when calculating density values.
    -------------------------    ---------------------------------------------------------
    radius_units                 Optional string. The units of the radius parameter. If no distance is provided, a default will be calculated that is based on the locations of the input features and the values in the count field (if a count field is provided).
                                 Choice list: ['Miles', 'Feet', 'Kilometers',  'Meters']
    -------------------------    ---------------------------------------------------------
    bounding_polygon_layer       Optional layer. A layer specifying the polygon(s) where you want densities to be calculated. For example, if you are interpolating densities of fish within a lake, you can use the boundary of the lake in this parameter and the output will only draw within the boundary of the lake. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    area_units                   Optional string. The units of the calculated density values.
                                 Choice list: ['areaUnits', 'SquareMiles']
    -------------------------    ---------------------------------------------------------
    classification_type          Optional string. Determines how density values will be classified into polygons.
                                 Choice list: ['EqualInterval', 'GeometricInterval', 'NaturalBreaks', 'EqualArea', 'StandardDeviation']
                                    * EqualInterval - Polygons are created such that the range of density values is equal for each area.
                                    * GeometricInterval - Polygons are based on class intervals that have a geometric series. This method ensures that each class range has approximately the same number of values within each class and that the change between intervals is consistent.
                                    * NaturalBreaks - Class intervals for polygons are based on natural groupings of the data. Class break values are identified that best group similar values and that maximize the differences between classes.
                                    * EqualArea - Polygons are created such that the size of each area is equal. For example, if the result has more high density values than low density values, more polygons will be created for high densities.
                                    * StandardDeviation - Polygons are created based upon the standard deviation of the predicted density values.
    -------------------------    ---------------------------------------------------------
    num_classes                  Optional int. This value is used to divide the range of predicted values into distinct classes. The range of values in each class is determined by the classification_type parameter.
    -------------------------    ---------------------------------------------------------
    output_name                  Optional string. Additional properties such as output feature service name.
    -------------------------    ---------------------------------------------------------
    context                      Optional string. Additional settings such as processing extent and output spatial reference. For calculate_density, there are two settings.

                                 #. Extent (extent)-a bounding box that defines the analysis area. Only those points in the input_layer that intersect the bounding box will be analyzed.
                                 #. Output Spatial Reference (outSR) the output features will be projected into the output spatial reference.
    -------------------------    ---------------------------------------------------------
    gis                          Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------    ---------------------------------------------------------
    estimate                     Optional Boolean. Is true, the number of credits needed to run the operation will be returned as a float.
    -------------------------    ---------------------------------------------------------
    future                       Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =========================    =========================================================


    :returns: result_layer : feature layer Item if output_name is specified, else Feature Collection.

    .. code-block:: python

        USAGE EXAMPLE: To create a layer that shows density of collisions within 2 miles.
                       The density is classified based upon the standard deviation.
                       The range of density values is divided into 5 classes.

        collision_density = calculate_density(input_layer=collisions,
                                        radius=2,
                                        radius_units='Miles',
                                        bounding_polygon_layer=zoning_lyr,
                                        area_units='SquareMiles',
                                        classification_type='StandardDeviation',
                                        num_classes=5,
                                        output_name='density_of_incidents')

    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.calculate_density(
        input_layer,
        field,
        cell_size,
        cell_size_units,
        radius,
        radius_units,
        bounding_polygon_layer,
        area_units,
        classification_type,
        num_classes,
        output_name,
        context,
        estimate=estimate, future=future)

def summarize_center_and_dispersion(
        analysis_layer,
        summarize_type,
        ellipse_size=None,
        weight_field=None,
        group_field=None,
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        future=False):

    """
    The Summarize Center and Dispersion task finds central features and directional distributions.

    ====================    =========================================================
    **Argument**            **Description**
    --------------------    ---------------------------------------------------------
    analysis_layer          The point, line, or polygon features to be analyzed. This
                            parameter can be a URL to a feature service layer with an
                            optional filter to select specific feaures, or a feature
                            collection
    --------------------    ---------------------------------------------------------
    summarize_type          The method with which to summarize the analysis_layer.
                            Choice List:
                            ["CentralFeature", "MeanCenter", "MedianCenter",
                            "Ellipse"]
                            Example: "CentralFeature"
    --------------------    ---------------------------------------------------------
    ellipse_size            The size of the output ellipse in standard deviations.
                            The default ellipse size is 1. Valid choices are 1, 2, or
                            3 standard deviations.
                            Choice List: [1, 2, 3]
                            Examples:
                            "1"
                            [1, 2, 3]
    --------------------    ---------------------------------------------------------
    weight_field            A numeric field in the analysis_layer to be used to
                            weight locations according to their relative importance.
    --------------------    ---------------------------------------------------------
    group_field             The field used to group features for separate directional
                            distribution calculations. The group_field can be of
                            integer, date, or string type.
    --------------------    ---------------------------------------------------------
    output_name             Optional string. Additional properties such as output
                            feature service name.
    --------------------    ---------------------------------------------------------
    context                 Optional string. Additional settings such as processing
                            extent and output spatial reference.
    --------------------    ---------------------------------------------------------
    gis                     Optional, the GIS on which this tool runs. If not
                            specified, the active GIS is used.
    --------------------    ---------------------------------------------------------
    estimate                Optional Boolean. If True, the number of credits to run the operation will be returned.
    --------------------    ---------------------------------------------------------
    future                  Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ====================    =========================================================

    :returns: Python dictionary with the following keys:
        "central_feature_result_layer" : layer (FeatureCollection)
        "mean_feature_result_layer" : layer (FeatureCollection)
        "median_feature_result_layer" : layer (FeatureCollection)
        "ellipse_feature_result_layer" : layer (FeatureCollection)
        "process_info" : list of messages
    """

    gis = _arcgis.env.active_gis if gis is None else gis

    return gis._tools.featureanalysis.summarize_center_and_dispersion(
        analysis_layer,
        summarize_type,
        ellipse_size,
        weight_field,
        group_field,
        output_name,
        context,
        estimate=estimate, future=future)

def find_point_clusters(
        analysis_layer,
        min_features_cluster,
        search_distance=None,
        search_distance_unit=None,
        output_name=None,
        context=None,
        gis=None, estimate=False, future=False):

    """
    .. image:: _static/images/find_point_clusters/find_point_clusters.png

    The ``find_point_clusters`` method finds clusters of point features within surrounding
    noise based on their spatial distribution.

    This method uses unsupervised machine learning clustering algorithms to detect
    patterns of point features based purely on spatial location and, optionally,
    the distance to a specified number of features.

    The result map shows each cluster identified as well as features considered
    noise. Multiple clusters will be assigned each color. Colors will be assigned
    and repeated so that each cluster is visually distinct from its neighboring clusters.

    This method utilizes two related algorithms. By default the HDBSCAN algorithm is
    used to find clusters. If a ``search_distance`` is specified, the DBSCAN algorithm
    is used. DBSCAN is only appropriate if there is a very clear search distance to use
    for your analysis and will return clusters with similar densities. When
    no ``search_distance`` is specified, HDBSCAN will use a range of distances to separate clusters
    of varying densities from sparser noise resulting in more data-driven clusters.

    ====================    =========================================================
    **Argument**            **Description**
    --------------------    ---------------------------------------------------------
    analysis_layer          Required layer. The point feature layer for which
                            density-based clustering will be calculated.
                            See :ref:`Feature Input<FeatureInput>`.
    --------------------    ---------------------------------------------------------
    min_features_cluster    Required integer. The minimum number of features to be
                            considered a cluster. Any cluster with fewer features
                            than the number provided will be considered noise.
    --------------------    ---------------------------------------------------------
    search_distance         Optional float. The maximum distance to consider. The
                            Minimum Features per Cluster specified must be found
                            within this distance for cluster membership. Individual
                            clusters will be separated by at least this distance. If
                            a feature is located further than this distance from the
                            next closest feature in the cluster, it will not be
                            included in the cluster.
    --------------------    ---------------------------------------------------------
    search_distance_unit    Optional string. The linear unit to be used with the distance
                            value specified for ``search_distance``. You must provide a
                            value if ``search_distance`` has been set.

                            Choice list: ['Feet', 'Miles', 'Meters', 'Kilometers']

                            The default is 'Miles'.
    --------------------    ---------------------------------------------------------
    output_name             Optional string. If provided, the method will create a
                            feature service of the results. You define the name of
                            the service. If ``output_name`` is not supplied, the method
                            will return a feature collection.
    --------------------    ---------------------------------------------------------
    context                 Optional string. Context contains additional settings that affect method execution. For ``find_point_clusters``, there are two settings.

                            #. Extent (``extent``) - a bounding box that defines the analysis area. Only those features in the input layer that intersect the bounding box will be buffered.
                            #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    --------------------    ---------------------------------------------------------
    gis                     Optional, the GIS on which this tool runs. If not
                            specified, the active GIS is used.
    --------------------    ---------------------------------------------------------
    estimate                Optional Boolean. If True, the number of credits to run the operation will be returned.
    --------------------    ---------------------------------------------------------
    future                  Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ====================    =========================================================

    :returns: result_layer : feature layer Item if ``output_name`` is specified, else Feature collection.

    .. code-block:: python

        USAGE EXAMPLE: To find patterns of taffic accidents purely on spatial location.
        clusters= find_point_clusters(collision,
                                      min_features_cluster=200,
                                      search_distance=2,
                                      search_distance_unit='Kilometers',
                                      output_name='find point clusters')

    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_point_clusters(
        analysis_layer,
        min_features_cluster,
        search_distance,
        search_distance_unit,
        output_name,
        context,
        estimate=estimate, future=future)

def find_hot_spots(
        analysis_layer,
        analysis_field=None,
        divided_by_field=None,
        bounding_polygon_layer=None,
        aggregation_polygon_layer=None,
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        shape_type=None,
        cell_size=None,
        cell_size_unit=None,
        distance_band=None,
        distance_band_unit=None,
        future=False):
    """
    .. image:: _static/images/find_hot_spots/find_hot_spots.png

    The ``find_hot_spots`` method analyzes point data (such as crime incidents, traffic accidents, or trees) or field values associated with
    points or area features (such as the number of people in each census tract or the total sales for retail stores). It finds statistically
    significant spatial clusters of high values (hot spots) and low values (cold spots). For point data when no field is specified, hot spots
    are locations with lots of points and cold spots are locations with very few points.

    The result map layer shows hot spots in red and cold spots in blue. The darkest red features indicate the strongest clustering of high values
    or point densities; you can be 99 percent confident that the clustering associated with these features could not be the result of random chance.
    Similarly, the darkest blue features are associated with the strongest spatial clustering of low values or the lowest point densities.
    Features that are beige are not part of a statistically significant cluster; the spatial pattern associated with these features could very likely
    be the result of random processes and random chance.

    ===================================================================    =========================================================
    **Argument**                                                           **Description**
    -------------------------------------------------------------------    ---------------------------------------------------------
    analysis_layer (Required if the analysis_layer contains polygons)      Required layer. The point or polygon feature layer for which hot spots will be calculated. See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------------------------------------    ---------------------------------------------------------
    analysis_field                                                         Optional string. The numeric field that will be analyzed. The field you select might represent:

                                                                            + counts (such as the number of traffic accidents)
                                                                            + rates (such as the number of crimes per square mile)
                                                                            + averages (such as the mean math test score)
                                                                            + indices (such as a customer satisfaction score)

                                                                           If an ``analysis_field`` is not supplied, hot spot results are based on point densities only.
    -------------------------------------------------------------------    ---------------------------------------------------------
    divided_by_field                                                       Optional string. The numeric field in the ``analysis_layer`` that will be used to normalize your data.
                                                                           For example, if your points represent crimes, dividing by total population would result in an analysis of crimes per capita rather than raw crime counts.

                                                                           You can use esriPopulation to geoenrich each area feature with the most recent population values, which will then be
                                                                           used as the attribute to divide by. This option will use credits.
    -------------------------------------------------------------------    ---------------------------------------------------------
    bounding_polygon_layer                                                 Optional layer. When the analysis layer is points and no ``analysis_field`` is specified, you can provide polygons features that define where incidents could have occurred.
                                                                           For example, if you are analyzing boating accidents in a harbor, the outline of the harbor might provide a good boundary for where accidents could occur.
                                                                           When no bounding areas are provided, only locations with at least one point will be included in the analysis. See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------------------------------------    ---------------------------------------------------------
    aggregation_polygon_layer                                              Optional layer. When the ``analysis_layer`` contains points and no ``analysis_field`` is specified,
                                                                           you can provide polygon features into which the points will be aggregated and analyzed, such as administrative units.
                                                                           The number of points that fall within each polygon are counted, and the point count in each polygon is analyzed. See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------------------------------------    ---------------------------------------------------------
    output_name                                                            Optional string. If provided, the task will create a feature service of the results.
                                                                           You define the name of the service. If ``output_name`` is not supplied, the task will return a feature collection.
    -------------------------------------------------------------------    ---------------------------------------------------------
    context                                                                Optional string. Context contains additional settings that affects method execution. For ``find_hot_spots``, there are two settings.

                                                                           #.  Extent (``extent``) - a bounding box that defines the analysis area. Only those features in the ``analysis_layer`` that intersect the bounding box will be analyzed.
                                                                           #. Output Spatial Reference (``outSR``) - the data will be projected into the output spatial reference prior to analysis.
    -------------------------------------------------------------------    ---------------------------------------------------------
    gis                                                                    Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------------------------------------------------    ---------------------------------------------------------
    estimate                                                               Optional Boolean. Is true, the number of credits needed to run the operation will be returned as a float.
    -------------------------------------------------------------------    ---------------------------------------------------------
    shape_type                                                             Optional string. The shape of the polygon mesh the input features will be aggregated into.

                                                                            * ``Fishnet``-The input features will be aggregated into a grid of square (fishnet) cells.
                                                                            * ``Hexagon``-The input features will be aggregated into a grid of hexagonal cells.
    -------------------------------------------------------------------    ---------------------------------------------------------
    cell_size                                                              Optional float. The size of the grid cells used to aggregate your features.
                                                                           When aggregating into a hexagon grid, this distance is used as the height to construct the hexagon polygons.
    -------------------------------------------------------------------    ---------------------------------------------------------
    cell_size_unit                                                         Optional string. The units of the ``cell_size`` value. You must provide a value if ``cell_size`` has been set.

                                                                           Choice list: ['Meters', 'Miles', 'Feet', 'Kilometers']
    -------------------------------------------------------------------    ---------------------------------------------------------
    distance_band                                                          Optional float. The spatial extent of the analysis neighborhood. This value determines which features are analyzed together
                                                                           in order to assess local clustering.
    -------------------------------------------------------------------    ---------------------------------------------------------
    distance_band_unit                                                     Optional string. The units of the ``distance_band`` value. You must provide a value if ``distance_band`` has been set.
    -------------------------------------------------------------------    ---------------------------------------------------------
    future                                                                 Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ===================================================================    =========================================================

    :returns: result_layer : feature layer Item if output_name is specified, else a dictionary with a Feature Collection and processing messages.

    .. code-block:: python

        USAGE EXAMPLE: To find significant hot ot cold spots of collisions involving a bicycle within a specific boundary.
        collision_hot_spots = find_hot_spots(collisions,
                                             bounding_polygon_layer=boundry_lyr,
                                             output_name='collision_hexagon_hot_spots',
                                             shape_type='hexagon')
    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_hot_spots(
        analysis_layer,
        analysis_field,
        divided_by_field,
        bounding_polygon_layer,
        aggregation_polygon_layer,
        output_name,
        context,
        estimate=estimate,
        shape_type=shape_type,
        cell_size=cell_size,
        cell_size_unit=cell_size_unit,
        distance_band=distance_band,
        distance_band_unit=distance_band_unit,
        future=future)


def find_outliers(analysis_layer,
                  analysis_field=None,
                  divided_by_field=None,
                  bounding_polygon_layer=None,
                  aggregation_polygon_layer=None,
                  permutations=None,
                  shape_type=None,
                  cell_size=None,
                  cell_units=None,
                  distance_band=None,
                  band_units=None,
                  output_name=None,
                  context=None,
                  gis=None,
                  estimate=False,
                  future=False):
    """
    .. image:: _static/images/find_outliers/find_outliers.png

    The ``find_outliers`` method analyzes point data (such as crime incidents, traffic accidents, or trees) or field values associated with points
    or area features (such as the number of people in each census tract or the total sales for retail stores). It finds statistically significant
    spatial clusters of high values and low values and statistically significant high or low spatial outliers within those clusters.

    The result map layer shows high outliers in red and low outliers in dark blue. Clusters of high values appear pink and clusters of low values
    appear light blue. Features that are beige are not a statistically significant outlier and not part of a statistically significant cluster; the
    spatial pattern associated with these features could very likely be the result of random processes and random chance.

    ==================================================================  ===============================================================
    **Argument**                                                        **Description**
    ------------------------------------------------------------------  ---------------------------------------------------------------
    analysis_layer                                                      Required feature layer. The point or polygon feature layer for which outliers will be calculated. See :ref:`Feature Input<FeatureInput>`.
    ------------------------------------------------------------------  ---------------------------------------------------------------
    analysis_field (Required if the analysis_layer contains polygons)   Optional string. The numeric field that will be analyzed. The field you select might represent:

                                                                        * counts (such as the number of traffic accidents)
                                                                        * rates (such as the number of crimes per square mile)
                                                                        * averages (such as the mean math test score)
                                                                        * indices (such as a customer satisfaction score)

                                                                        If an ``analysis_field`` is not supplied, hot spot results are based on point densities only.
    ------------------------------------------------------------------  ---------------------------------------------------------------
    divided_by_field                                                    Optional string. The numeric field in the ``analysis_layer`` that will be used to normalize your data.
                                                                        For example, if your points represent crimes, dividing by total population would result in an analysis
                                                                        of crimes per capita rather than raw crime counts.

                                                                        You can use esriPopulation to geoenrich each area feature with the most recent population values,
                                                                        which will then be used as the attribute to divide by. This option will use credits.
    ------------------------------------------------------------------  ---------------------------------------------------------------
    bounding_polygon_layer                                              Optional layer. When the analysis layer is points and no ``analysis_field`` is specified, you can provide polygon features that define where incidents could have occurred.
                                                                        For example, if you are analyzing boating accidents in a harbor, the outline of the harbor might provide a good boundary for where accidents could occur.
                                                                        When no bounding areas are provided, only locations with at least one point will be included in the analysis. See :ref:`Feature Input<FeatureInput>`.
    ------------------------------------------------------------------  ---------------------------------------------------------------
    aggregation_polygon_layer                                           Optional layer. When the ``analysis_layer`` contains points and no ``analysis_field`` is specified, you can provide polygon features into which the
                                                                        points will be aggregated and analyzed, such as administrative units. The number of points that fall within each polygon
                                                                        are counted, and the point count in each polygon is analyzed. See :ref:`Feature Input<FeatureInput>`.
    ------------------------------------------------------------------  ---------------------------------------------------------------
    permutations                                                        Optional string. Permutations are used to determine how likely it would be to find the actual spatial distribution of the values you are analyzing.
                                                                        Choosing the number of permutations is a balance between precision and increased processing time. A lower number of permutations
                                                                        can be used when first exploring a problem, but it is best practice to increase the permutations to the highest number feasible for final results.

                                                                        Choice list: ['Speed', 'Balance', 'Presision']

                                                                        * ``Speed`` - implements 199 permutations and results in p-values with a precision of 0.01.
                                                                        * ``Balance`` - implements 499 permutations and results in p-values with a precision of 0.002.
                                                                        * ``Precision`` - implements 999 permutations and results in p-values with a precision of 0.001.
    ------------------------------------------------------------------  ---------------------------------------------------------------
    shape_type                                                          Optional string. The shape of the polygon mesh the input features will be aggregated into.

                                                                        * ``Fishnet`` - The input features will be aggregated into a grid of square (fishnet) cells.
                                                                        * ``Hexagon`` -  The input features will be aggregated into a grid of hexagonal cells.
    ------------------------------------------------------------------  ---------------------------------------------------------------
    cell_size                                                           Optional float. The size of the grid cells used to aggregate your features. When aggregating into a hexagon grid, this distance is used as the height to construct the hexagon polygons.
    ------------------------------------------------------------------  ---------------------------------------------------------------
    cell_units                                                          Optional string. The units of the ``cell_size`` value. You must provide a value if ``cell_size`` has been set.

                                                                        Choice list: ['Meters', 'Miles', 'Feet', 'Kilometers']
    ------------------------------------------------------------------  ---------------------------------------------------------------
    distance_band                                                       Optional float. The spatial extent of the analysis neighborhood. This value determines which features are analyzed together
                                                                        in order to assess local clustering.
    ------------------------------------------------------------------  ---------------------------------------------------------------
    band_units                                                          Optional string. The units of the ``distance_band`` value. You must provide a value if ``distance_band`` has been set.

                                                                        Choice list: ['Meters', 'Miles', 'Feet', 'Kilometers']
    ------------------------------------------------------------------  ---------------------------------------------------------------
    output_name                                                         Optional string. If provided, the method will create a feature service of the results. You define the name of the service. If ``output_name`` is not supplied, the method will return a feature collection.
    ------------------------------------------------------------------  ---------------------------------------------------------------
    context                                                             Context contains additional settings that affect task execution. For ``find_outliers``, there are two settings:

                                                                        #. Extent (extent) a bounding box that defines the analysis area. Only those features in the ``analysis_layer`` that intersect the bounding box will be analyzed.

                                                                        #. Output Spatial Reference (outSR) the data will be projected into the output spatial reference prior to analysis.
    ------------------------------------------------------------------  ---------------------------------------------------------------
    estimate                                                            Optional boolean. Returns the number of credit for the operation.
    ------------------------------------------------------------------  ---------------------------------------------------------------
    future                                                              Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ==================================================================  ===============================================================

    :Returns:
    Item if output_name is set. else results in a dict with the following keys:

       "find_outliers_result_layer" : layer (FeatureCollection)

       "process_info" : list of messages
    .. code-block:: python

         #USAGE EXAMPLE: To find statistically significant outliers within the collision clusters.
         outliers = find_outliers(analysis_layer=collisions,
                                  shape_type='fishnet',
                                  output_name='find outliers')

    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_outliers(analysis_layer,
                                                    analysis_field,
                                                    divided_by_field,
                                                    bounding_polygon_layer,
                                                    aggregation_polygon_layer,
                                                    permutations,
                                                    shape_type,
                                                    cell_size,
                                                    cell_units,
                                                    distance_band,
                                                    band_units,
                                                    output_name,
                                                    context,
                                                    estimate=estimate,
                                                    future=future)


def interpolate_points(
        input_layer,
        field,
        interpolate_option="5",
        output_prediction_error=False,
        classification_type="GeometricInterval",
        num_classes=10,
        class_breaks=[],
        bounding_polygon_layer=None,
        predict_at_point_layer=None,
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        future=False):
    """
    .. image:: _static/images/interpolate_points/interpolate_points.png

    The ``interpolate_points`` method allows you to predict values at new locations based on measurements
    from a collection of points. The method takes point data with values at each point and returns
    areas classified by predicted values. For example:

    * An air quality management district has sensors that measure pollution levels.
      ``interpolate_points`` can be used to predict pollution levels at locations that don't have sensors,
      such as locations with at-risk populations, schools, or hospitals, for example.
    * Predict heavy metal concentrations in crops based on samples taken from individual plants.
    * Predict soil nutrient levels (nitrogen, phosphorus, potassium, and so on) and other
      indicators (such as electrical conductivity) in order to study their relationships to crop yield
      and prescribe precise amounts of fertilizer for each location in the field.
    * Meteorological applications include prediction of temperatures, rainfall,
      and associated variables (such as acid rain).

    ``interpolate_points`` uses the `Empirical Bayesian
    Kriging <http://desktop.arcgis.com/en/arcmap/latest/tools/geostatistical-analyst-toolbox/empirical-bayesian-kriging.htm>`_
    geoprocessing tool to perform the interpolation. The parameters that are supplied to
    the Empirical Bayesian Kriging tool are controlled by the ``interpolate_option`` request parameter.

    If a value of 1 is provided for ``interpolate_option``, empirical Bayesian kriging will
    use the following parameters:

    * transformation_type - NONE
    * semivariogram_model_type - POWER
    * max_local_points - 50
    * overlap_factor - 1
    * number_semivariograms - 30
    * nbrMin - 8
    * nbrMax - 8

    If a value of 5 is provided for ``interpolate_option``, empirical Bayesian kriging
    will use the following parameters:

    * transformation_type - NONE
    * semivariogram_model_type - POWER
    * max_local_points 75
    * overlap_factor - 1.5
    * number_semivariograms - 100
    * nbrMin - 10
    * nbrMax - 10

    If a value of 9 is provided for ``interpolate_option``, empirical Bayesian kriging
    will use the following parameters:

    * transformation_type - EMPIRICAL
    * semivariogram_model_type - K_BESSEL
    * max_local_points - 200
    * overlap_factor - 3
    * number_semivariograms - 200
    * nbrMin - 15
    * nbrMax - 15

    ===========================  ===========================================================================================
    **Argument**                 **Description**
    ---------------------------  -------------------------------------------------------------------------------------------
    input_layer                  Required layer. The point layer whose features will be interpolated. See :ref:`Feature Input<FeatureInput>`.
    ---------------------------  -------------------------------------------------------------------------------------------
    field                        Required string. Name of the numeric field containing the values you wish to interpolate.
    ---------------------------  -------------------------------------------------------------------------------------------
    interpolate_option           Optional integer. Integer value declaring your preference for speed versus accuracy, from 1 (fastest) to 9 (most accurate).
                                 More accurate predictions take longer to calculate.

                                 Choice list: [1, 5, 9].

                                 The default is 5.
    ---------------------------  -------------------------------------------------------------------------------------------
    output_prediction_error      Optional boolean. If True, a polygon layer of standard errors for the interpolation
                                 predictions will be returned in the ``prediction_error`` output parameter.

                                 Standard errors are useful because they provide information about the reliability of the predicted values.
                                 A simple rule of thumb is that the true value will fall within two standard errors of the predicted
                                 value 95 percent of the time. For example, suppose a new location gets a predicted value of 50 with a
                                 standard error of 5. This means that this task's best guess is that the true value at that location is 50,
                                 but it reasonably could be as low as 40 or as high as 60. To calculate this range of reasonable values,
                                 multiply the standard error by 2, add this value to the predicted value to get the upper end of the range,
                                 and subtract it from the predicted value to get the lower end of the range.
    ---------------------------  -------------------------------------------------------------------------------------------
    classification_type          Optional string. Determines how predicted values will be classified into areas.

                                 * ``EqualArea`` - Polygons are created such that the number of data values in each area is equal.
                                   For example, if the data has more large values than small values, more areas will be created for large values.
                                 * ``EqualInterval`` - Polygons are created such that the range of predicted values is equal for each area.
                                 * ``GeometricInterval`` - Polygons are based on class intervals that have a geometrical series.
                                   This method ensures that each class range has approximately the same number of values within
                                   each class and that the change between intervals is consistent.
                                 * ``Manual`` - You to define your own range of values for areas. These values will be entered in
                                   the ``class_breaks`` parameter below.

                                 Choice list: ['EqualArea', 'EqualInterval', 'GeometricInterval', 'Manual']

                                 The default is 'GeometricInterval'.
    ---------------------------  -------------------------------------------------------------------------------------------
    num_classes                  Optional integer. This value is used to divide the range of interpolated values into distinct classes.
                                 The range of values in each class is determined by the ``classification_type`` parameter.
                                 Each class defines the boundaries of the result polygons.

                                 The default is 10. The maximum value is 32.
    ---------------------------  -------------------------------------------------------------------------------------------
    class_breaks                 Optional list of floats. If ``classification_type`` is Manual, supply desired class break values separated by spaces.
                                 These values define the upper limit of each class, so the number of classes will equal the number of entered values.
                                 Areas will not be created for any locations with predicted values above the largest entered break value.
                                 You must enter at least two values and no more than 32.
    ---------------------------  -------------------------------------------------------------------------------------------
    bounding_polygon_layer       Optional layer. A layer specifying the polygon(s) where you want values to be interpolated.  For example,
                                 if you are interpolating densities of fish within a lake, you can use the boundary of the lake in this
                                 parameter and the output will only contain polygons within the boundary of the lake. See :ref:`Feature Input<FeatureInput>`.
    ---------------------------  -------------------------------------------------------------------------------------------
    predict_at_point_layer       Optional layer. An optional layer specifying point locations to calculate prediction values.
                                 This allows you to make predictions at specific locations of interest. For example, if the ``input_layer`` represents
                                 measurements of pollution levels, you can use this parameter to predict the pollution levels of locations with large
                                 at-risk populations, such as schools or hospitals. You can then use this information to give recommendations to health
                                 officials in those locations.

                                 If supplied, the output ``predicted_point_layer`` will contain predictions at the specified locations. See :ref:`Feature Input<FeatureInput>`.
    ---------------------------  -------------------------------------------------------------------------------------------
    output_name                  Optional string. If provided, the method will create a feature service of the results.
                                 You define the name of the service. If ``output_name`` is not supplied, the method will return a feature collection.
    ---------------------------  -------------------------------------------------------------------------------------------
    context                      Optional string. Additional settings such as processing extent and output spatial reference.
    ---------------------------  -------------------------------------------------------------------------------------------
    gis                          Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    ---------------------------  -------------------------------------------------------------------------------------------
    estimate                     Optional boolean. If True, the number of credits to run the operation will be returned.
    ---------------------------  -------------------------------------------------------------------------------------------
    future                       Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ===========================  ===========================================================================================

    :returns: result_layer : feature layer Item if ``output_name`` is specified, else Python dictionary with the following keys:

        "result_layer" : layer (FeatureCollection)

        "prediction_error" : layer (FeatureCollection)

        "predicted_point_layer" : layer (FeatureCollection)

    .. code-block:: python

        #USAGE EXAMPLE: To predict mine production in US at new locations.
        interpolated = interpolate_points(coal_mines_us,
                                          field='Total_Prod',
                                          interpolate_option=5,
                                          output_prediction_error=True,
                                          classification_type='GeometricInterval',
                                          num_classes=10,
                                          output_name='interpolate coal mines production')
    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.interpolate_points(
        input_layer,
        field,
        interpolate_option,
        output_prediction_error,
        classification_type,
        num_classes,
        class_breaks,
        bounding_polygon_layer,
        predict_at_point_layer,
        output_name,
        context,
        estimate=estimate,
        future=future)




import os
from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap

class SiteManager(object):
    """
    Provides the ability to update and restore notebook sites.

    """
    _nb = None
    _url = None
    _gis = None
    _con = None
    _properties = None


    #----------------------------------------------------------------------
    def __init__(self, url, notebook, gis):
        """Constructor"""
        self._url = url
        self._nb = notebook
        gis = notebook._gis
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._con
        else:
            raise ValueError("Invalid GIS object")
    #----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {'f': 'json'}
            res = self._gis._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def __str__(self):
        return "<SiteManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return "<SiteManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties of the resource"""
        if self._properties is None:
            self._init()
        return self._properties
    #----------------------------------------------------------------------
    def export_site(self, location):
        """
        ArcGIS Notebook Server provides this operation to back up the site's
        configuration store, along with the importSite operation to restore
        a site configuration from a backup. The configuration store hosts
        essential information about the ArcGIS Notebook Server site and its
        machines.

        The output of this operation is a ZIP file with the .agssite file
        extension.

        There are many items and directories that are not backed up by this
        operation. Among them:

        - Notebooks
        - Container settings
        - Jobs directory

        If desired, you can create your own file system backups for these items.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        location               Require String. The folder to save the site to.
        ==================     ====================================================================

        :returns: string

        """
        url = "{base}/exportSite".format(base=self._url)
        params = {'f' : 'json'}
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def import_site(self,
                    location):
        """
        ArcGIS Notebook Server provides this operation to restore a site
        configuration from a backup. The backup will have been created and
        exported by the exportSite operation as a ZIP file with the
        .agssite file extension.

        Performing this operation will overwrite the current contents of
        your ArcGIS Notebook Server site's configuration store with the
        contents from the backup. You can use it to restore a site
        configuration in the event of machine failure or human error.

        The import operation may take a while to complete. When you execute
        the operation, keep the tab open on your browser until the
        operation completes, as a report will be delivered to the page.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        location               Require String. Path to the backup site file.
        ==================     ====================================================================

        :returns: Boolean

        """
        url = "{base}/importSite".format(base=self._url)
        params = {'f' : 'json'}
        params['EXPORT_LOCATION'] = location
        res = self._con.post(url, params)
        if 'status' in res:
            return res['status']
        return res
